"""Module for fetching and processing NWM gridded data."""
from typing import Union, List, Optional, Dict, Annotated
from datetime import datetime
from pathlib import Path

from pydantic import validate_call, Field, InstanceOf
from geopandas import GeoDataFrame

from teehr.fetching.nwm.grid_utils import fetch_and_format_nwm_grids
from teehr.fetching.utils import (
    build_remote_nwm_filelist,
    generate_json_paths,
    validate_operational_start_end_date,
    start_on_z_hour,
    end_on_z_hour,
    get_dataset
)
from teehr.models.fetching.utils import (
    SupportedNWMOperationalVersionsEnum,
    SupportedNWMDataSourcesEnum,
    SupportedKerchunkMethod,
    TimeseriesTypeEnum
)
from teehr.fetching.const import (
    NWM12_ANALYSIS_CONFIG,
    NWM20_ANALYSIS_CONFIG,
    NWM22_ANALYSIS_CONFIG,
    NWM30_ANALYSIS_CONFIG,
)
from teehr.utilities.generate_weights import generate_weights_file


@validate_call()
def nwm_grids_to_parquet(
    configuration: str,
    output_type: str,
    variable_name: str,
    start_date: Union[str, datetime],
    ingest_days: int,
    zonal_weights_filepath: Union[Path, str],
    json_dir: Union[str, Path],
    output_parquet_dir: Union[str, Path],
    nwm_version: SupportedNWMOperationalVersionsEnum,
    data_source: Optional[SupportedNWMDataSourcesEnum] = "GCS",
    kerchunk_method: Optional[SupportedKerchunkMethod] = "local",
    prioritize_analysis_valid_time: Optional[bool] = False,
    t_minus_hours: Optional[List[int]] = None,
    ignore_missing_file: Optional[bool] = True,
    overwrite_output: Optional[bool] = False,
    location_id_prefix: Optional[Union[str, None]] = None,
    variable_mapper: Dict[str, Dict[str, str]] = None,
    timeseries_type: TimeseriesTypeEnum = "primary",
    starting_z_hour: Optional[Annotated[int, Field(ge=0, le=23)]] = None,
    ending_z_hour: Optional[Annotated[int, Field(ge=0, le=23)]] = None,
    calculate_zonal_weights: bool = False,
    zone_polygons: Optional[Union[Path, str, InstanceOf[GeoDataFrame]]] = None,
    unique_zone_id: Optional[str] = None,
):
    """
    Fetch NWM gridded data, calculate zonal statistics (currently only
    mean is available) of selected variable for given zones, convert
    and save to TEEHR tabular format.

    Parameters
    ----------
    configuration : str
        NWM forecast category.
        (e.g., "analysis_assim", "short_range", ...).
    output_type : str
        Output component of the configuration.
        (e.g., "channel_rt", "reservoir", ...).
    variable_name : str
        Name of the NWM data variable to download.
        (e.g., "streamflow", "velocity", ...).
    start_date : str or datetime
        Date to begin data ingest.
        Str formats can include YYYY-MM-DD or MM/DD/YYYY.
    ingest_days : int
        Number of days to ingest data after start date.
    zonal_weights_filepath : str
        Path to the array containing fraction of pixel overlap
        for each zone.
    json_dir : str
        Directory path for saving json reference files.
    output_parquet_dir : str
        Path to the directory for the final parquet files.
    nwm_version : SupportedNWMOperationalVersionsEnum
        The NWM operational version.
        "nwm12", "nwm20", "nwm21", "nwm22", or "nwm30".
        Note that there is no change in NWM configuration between
        version 2.1 and 2.2, and they are treated as the same version.
        They are both allowed here for convenience.

        Availability of each version:

        - v1.2: 2018-09-17 - 2019-06-18
        - v2.0: 2019-06-19 - 2021-04-19
        - v2.1/2.2: 2021-04-20 - 2023-09-18
        - v3.0: 2023-09-19 - present
    data_source : Optional[SupportedNWMDataSourcesEnum]
        Specifies the remote location from which to fetch the data
        "GCS" (default), "NOMADS", or "DSTOR".
        Currently only "GCS" is implemented.
    kerchunk_method : Optional[SupportedKerchunkMethod]
        When data_source = "GCS", specifies the preference in creating Kerchunk
        reference json files. "local" (default) will create new json files from
        netcdf files in GCS and save to a local directory if they do not already
        exist locally, in which case the creation is skipped. "remote" - read the
        CIROH pre-generated jsons from s3, ignoring any that are unavailable.
        "auto" - read the CIROH pre-generated jsons from s3, and create any that
        are unavailable, storing locally.
    prioritize_analysis_valid_time : Optional[bool]
        A boolean flag that determines the method of fetching analysis data.
        When False (default), all hours of the reference time are included in the
        output. When True, only the hours within t_minus_hours are included.
    t_minus_hours : Optional[Iterable[int]]
        Specifies the look-back hours to include if an assimilation
        configuration is specified.
    ignore_missing_file : bool
        Flag specifying whether or not to fail if a missing NWM file is encountered
        True = skip and continue; False = fail.
    overwrite_output : bool
        Flag specifying whether or not to overwrite output files if they already
        exist.  True = overwrite; False = fail.
    location_id_prefix : Union[str, None]
        Optional location ID prefix to add (prepend) or replace.
    starting_z_hour : Optional[int]
        The starting z_hour to include in the output. If None, all z_hours
        are included for the first day. Default is None. Must be between 0 and 23.
    ending_z_hour : Optional[int]
        The ending z_hour to include in the output. If None, all z_hours
        are included for the last day. Default is None. Must be between 0 and 23.
    variable_mapper : Dict[str, Dict[str, str]]
        A dictionary of dictionaries to map NWM variable names to new names.
    timeseries_type : TimeseriesTypeEnum
        The type of timeseries to generate.
        "primary" (default) or "secondary".
    calculate_zonal_weights : bool
        Flag to calculate zonal weights.
    zone_polygons : Union[Path, str, InstanceOf[GeoDataFrame]]
        Path to the polygons file or a GeoDataFrame.
    unique_zone_id : Optional[str]
        Name of the field in the zone polygon file containing unique IDs.


    See Also
    --------
    teehr.utilities.generate_weights.generate_weights_file : Weighted average.

    Notes
    -----
    The NWM configuration variables, including configuration, output_type, and
    variable_name are stored as a pydantic model in grid_config_models.py.

    Forecast and assimilation data is grouped and saved one file per reference
    time, using the file name convention "YYYYMMDDTHH".  The tabular output
    parquet files follow the timeseries data model described in the
    :ref:`data model <data_model>`.

    Additionally, the location_id values in the zonal weights file are used as
    location ids in the output of this function, unless a prefix is specified which
    will be prepended to the location_id values if none exists, or will it replace
    the existing prefix. It is assumed that the location_id follows the pattern
    '[prefix]-[unique id]'.

    All dates and times within the files and in the file names are in UTC.

    Examples
    --------
    Here we will calculate mean areal precipitation using NWM forcing data for
    some watersheds (polygons) a using pre-calculated weights file
    (see: :func:`generate_weights_file()
    <teehr.utilities.generate_weights.generate_weights_file>` for weights calculation).

    Import the necessary module.

    >>> import teehr.fetching.nwm.nwm_grids as tlg

    Specify the input variables.

    >>> CONFIGURATION = "forcing_short_range"
    >>> OUTPUT_TYPE = "forcing"
    >>> VARIABLE_NAME = "RAINRATE"
    >>> START_DATE = "2020-12-18"
    >>> INGEST_DAYS = 1
    >>> ZONAL_WEIGHTS_FILEPATH = Path(Path.home(), "nextgen_03S_weights.parquet")
    >>> JSON_DIR = Path(Path.home(), "temp/parquet/jsons/")
    >>> OUTPUT_DIR = Path(Path.home(), "temp/parquet")
    >>> NWM_VERSION = "nwm22"
    >>> DATA_SOURCE = "GCS"
    >>> KERCHUNK_METHOD = "auto"
    >>> T_MINUS = [0, 1, 2]
    >>> IGNORE_MISSING_FILE = True
    >>> OVERWRITE_OUTPUT = True

    Perform the calculations, writing to the specified directory.

    >>> tlg.nwm_grids_to_parquet(
    >>>     nwm_configuration=CONFIGURATION,
    >>>     output_type=OUTPUT_TYPE,
    >>>     variable_name=VARIABLE_NAME,
    >>>     start_date=START_DATE,
    >>>     ingest_days=INGEST_DAYS,
    >>>     zonal_weights_filepath=ZONAL_WEIGHTS_FILEPATH,
    >>>     json_dir=JSON_DIR,
    >>>     output_parquet_dir=OUTPUT_DIR,
    >>>     nwm_version=NWM_VERSION,
    >>>     data_source=DATA_SOURCE,
    >>>     kerchunk_method=KERCHUNK_METHOD,
    >>>     t_minus_hours=T_MINUS,
    >>>     ignore_missing_file=IGNORE_MISSING_FILE,
    >>>     overwrite_output=OVERWRITE_OUTPUT
    >>> )
    """ # noqa
    # Import appropriate config model and dicts based on NWM version
    if nwm_version == SupportedNWMOperationalVersionsEnum.nwm12:
        from teehr.models.fetching.nwm12_grid import GridConfigurationModel
        analysis_config_dict = NWM12_ANALYSIS_CONFIG
    elif nwm_version == SupportedNWMOperationalVersionsEnum.nwm20:
        from teehr.models.fetching.nwm20_grid import GridConfigurationModel
        analysis_config_dict = NWM20_ANALYSIS_CONFIG
    elif nwm_version == SupportedNWMOperationalVersionsEnum.nwm21:
        from teehr.models.fetching.nwm22_grid import GridConfigurationModel
        analysis_config_dict = NWM22_ANALYSIS_CONFIG
    elif nwm_version == SupportedNWMOperationalVersionsEnum.nwm22:
        from teehr.models.fetching.nwm22_grid import GridConfigurationModel
        analysis_config_dict = NWM22_ANALYSIS_CONFIG
    elif nwm_version == SupportedNWMOperationalVersionsEnum.nwm30:
        from teehr.models.fetching.nwm30_grid import GridConfigurationModel
        analysis_config_dict = NWM30_ANALYSIS_CONFIG
    else:
        raise ValueError(
            "nwm_version must equal "
            "'nwm12', 'nwm20', 'nwm21', 'nwm22' or 'nwm30'"
        )

    # Parse input parameters to validate configuration
    vars = {
        "configuration": configuration,
        configuration: {
            "output_type": output_type,
            output_type: variable_name,
        },
    }

    cm = GridConfigurationModel.model_validate(vars)
    configuration = cm.configuration.name
    forecast_obj = getattr(cm, configuration)
    output_type = forecast_obj.output_type.name
    variable_name = getattr(forecast_obj, output_type).name

    # Check data_source
    if data_source == SupportedNWMDataSourcesEnum.NOMADS:
        # TODO
        raise ValueError("Fetching from NOMADS is not yet implemented")
    elif data_source == SupportedNWMDataSourcesEnum.DSTOR:
        # TODO
        raise ValueError("Fetching from DSTOR is not yet implemented")
    else:

        # Make sure start/end dates work with specified NWM version
        validate_operational_start_end_date(
            nwm_version,
            start_date,
            ingest_days
        )

        # Build paths to netcdf files on GCS
        gcs_component_paths = build_remote_nwm_filelist(
            configuration,
            output_type,
            start_date,
            ingest_days,
            analysis_config_dict,
            t_minus_hours,
            ignore_missing_file,
            prioritize_analysis_valid_time
        )

        if starting_z_hour is not None:
            gcs_component_paths = start_on_z_hour(
                start_date=start_date,
                start_z_hour=starting_z_hour,
                gcs_component_paths=gcs_component_paths
            )

        if ending_z_hour is not None:
            gcs_component_paths = end_on_z_hour(
                start_date=start_date,
                ingest_days=ingest_days,
                end_z_hour=ending_z_hour,
                gcs_component_paths=gcs_component_paths
            )

        # Create paths to local and/or remote kerchunk jsons
        json_paths = generate_json_paths(
            kerchunk_method,
            gcs_component_paths,
            json_dir,
            ignore_missing_file
        )

        # If specified, generate zonal weights file here.
        if calculate_zonal_weights:
            if zone_polygons is None:
                raise ValueError(
                    "The zone polygons must be provided"
                    " to calculate zonal weights. Can be a GeoDataFame"
                    " or a filepath."
                )

            # Get a single timestep to use as a template grid.
            template_ds = get_dataset(
                json_paths[0],
                ignore_missing_file=False,
                target_options={'anon': True}
            )

            generate_weights_file(
                zone_polygons=zone_polygons,
                template_dataset=template_ds,
                variable_name=variable_name,
                crs_wkt=template_ds.crs.esri_pe_string,
                output_weights_filepath=zonal_weights_filepath,
                location_id_prefix=location_id_prefix,
                unique_zone_id=unique_zone_id
            )

        # Fetch the data, saving to parquet files based on TEEHR data model
        fetch_and_format_nwm_grids(
            json_paths=json_paths,
            nwm_configuration_name=configuration,
            nwm_version=nwm_version,
            variable_name=variable_name,
            output_parquet_dir=output_parquet_dir,
            zonal_weights_filepath=zonal_weights_filepath,
            ignore_missing_file=ignore_missing_file,
            overwrite_output=overwrite_output,
            location_id_prefix=location_id_prefix,
            variable_mapper=variable_mapper,
            timeseries_type=timeseries_type
        )


"""Utility function for setting up the NWM streamflow fetching example."""
from pathlib import Path
import shutil

import teehr
import teehr.example_data.nwm_gridded_example_data as fetch_nwm_grid_data


def setup_nwm_example(tmpdir):
    """Set up the NWM streamflow fetching example."""
    # Create an Evaluation object and create the directory
    ev = teehr.Evaluation(dir_path=tmpdir, create_dir=True)

    # Clone the template
    ev.clone_template()

    # Fetch the test data
    location_data_path = Path(tmpdir, "usgs_at_radford_location.parquet")
    fetch_nwm_grid_data.fetch_file("usgs_at_radford_location.parquet", location_data_path)

    wbd_location_data_path = Path(tmpdir, "three_huc10s_radford.parquet")
    fetch_nwm_grid_data.fetch_file("three_huc10s_radford.parquet", wbd_location_data_path)

    crosswalk_data_path = Path(tmpdir, "location_crosswalks.parquet")
    fetch_nwm_grid_data.fetch_file("location_crosswalks.parquet", crosswalk_data_path)

    forcing_crosswalk_data_path = Path(tmpdir, "forcing_xwalk.parquet")
    fetch_nwm_grid_data.fetch_file("forcing_xwalk.parquet", forcing_crosswalk_data_path)

    primary_timeseries_path = Path(tmpdir, "primary_timeseries.parquet")
    fetch_nwm_grid_data.fetch_file("primary_timeseries.parquet", primary_timeseries_path)

    secondary_timeseries_path = Path(tmpdir, "secondary_timeseries.parquet")
    fetch_nwm_grid_data.fetch_file("secondary_timeseries.parquet", secondary_timeseries_path)

    joined_timeseries_path = Path(tmpdir, "joined_timeseries.parquet")
    fetch_nwm_grid_data.fetch_file("joined_timeseries.parquet", joined_timeseries_path)

    configurations_path = Path(tmpdir, "configurations.parquet")
    fetch_nwm_grid_data.fetch_file("configurations.parquet", configurations_path)

    weights_path = Path(tmpdir, "nwm30_forcing_analysis_assim_pixel_weights.parquet")
    fetch_nwm_grid_data.fetch_file("nwm30_forcing_analysis_assim_pixel_weights.parquet", weights_path)

    # Manually load the data into the Evaluation
    # joined timeseries
    shutil.move(
        src=joined_timeseries_path,
        dst=ev.joined_timeseries.dir
    )
    # configurations
    shutil.move(
        src=configurations_path,
        dst=ev.configurations.dir
    )
    # secondary
    shutil.move(
        src=secondary_timeseries_path,
        dst=ev.secondary_timeseries.dir
    )
    # primary
    shutil.move(
        src=primary_timeseries_path,
        dst=ev.primary_timeseries.dir
    )
    # location crosswalks
    shutil.move(
        src=crosswalk_data_path,
        dst=ev.location_crosswalks.dir
    )
    # locations
    shutil.move(
        src=location_data_path,
        dst=ev.locations.dir
    )

    # weights file
    Path(ev.cache_dir, "fetching", "weights", "nwm30_forcing_analysis_assim").mkdir(parents=True)
    shutil.move(
        src=weights_path,
        dst=Path(ev.cache_dir, "fetching", "weights", "nwm30_forcing_analysis_assim")
    )


if __name__ == "__main__":

    # Define the directory where the Evaluation will be created.
    test_eval_dir = Path(Path().home(), "temp", "11_fetch_nwm_gridded_data")
    shutil.rmtree(test_eval_dir, ignore_errors=True)

    # Setup the example evaluation using data from the TEEHR repository.
    setup_nwm_example(tmpdir=test_eval_dir)

    # Initialize the evaluation.
    ev = teehr.Evaluation(dir_path=test_eval_dir)

    pass


"""Module defining shared functions for processing NWM point data."""
from pathlib import Path
from typing import Dict, Iterable, List, Tuple
import re

import dask
import numpy as np
import pandas as pd
import pyarrow as pa

from teehr.fetching.utils import (
    get_dataset,
    write_timeseries_parquet_file,
    split_dataframe,
    format_nwm_configuration_name,
    parse_nwm_json_paths
)
from teehr.models.fetching.utils import TimeseriesTypeEnum
from teehr.fetching.const import (
    VALUE,
    VALUE_TIME,
    REFERENCE_TIME,
    LOCATION_ID,
    UNIT_NAME,
    VARIABLE_NAME,
    CONFIGURATION_NAME,
    MEMBER
)


@dask.delayed
def file_chunk_loop(
    row: Tuple,
    location_ids: np.array,
    variable_name: str,
    configuration: str,
    schema: pa.Schema,
    ignore_missing_file: bool,
    nwm_version: str,
    variable_mapper: Dict[str, Dict[str, str]]
):
    """Fetch NWM values and convert to tabular format for a single json."""
    ds = get_dataset(
        row.filepath,
        ignore_missing_file,
        target_options={'anon': True}
    )
    if not ds:
        return None
    ds = ds.sel(feature_id=location_ids)
    vals = ds[variable_name].astype("float32").values

    nwm_units = ds[variable_name].units

    if not variable_mapper:
        teehr_units = nwm_units
        teehr_variable_name = variable_name
    else:
        teehr_units = variable_mapper[UNIT_NAME].get(nwm_units, nwm_units)
        teehr_variable_name = variable_mapper[VARIABLE_NAME].get(
            variable_name, variable_name
        )

    ref_time = pd.to_datetime(row.day) \
        + pd.to_timedelta(int(row.z_hour[1:3]), unit="h")

    valid_time = ds.time.values
    feature_ids = ds.feature_id.astype("int32").values
    teehr_location_ids = \
        [f"{nwm_version}-{feat_id}" for feat_id in feature_ids]
    num_vals = vals.size

    teehr_config = format_nwm_configuration_name(
        nwm_configuration_name=configuration,
        nwm_version=nwm_version
    )

    output_table = pa.table(
        {
            VALUE: vals,
            REFERENCE_TIME: np.full(vals.shape, ref_time),
            LOCATION_ID: teehr_location_ids,
            VALUE_TIME: np.full(vals.shape, valid_time),
            CONFIGURATION_NAME: num_vals * [teehr_config["configuration_name"]],
            VARIABLE_NAME: num_vals * [teehr_variable_name],
            UNIT_NAME: num_vals * [teehr_units],
            MEMBER: num_vals * [teehr_config["member"]]
        },
        schema=schema,
    )

    return output_table


def process_chunk_of_files(
    df: pd.DataFrame,
    location_ids: Iterable[int],
    configuration: str,
    variable_name: str,
    output_parquet_dir: str,
    process_by_z_hour: bool,
    ignore_missing_file: bool,
    overwrite_output: bool,
    nwm_version: str,
    variable_mapper: Dict[str, Dict[str, str]],
    timeseries_type: TimeseriesTypeEnum
):
    """Assemble a table for a chunk of NWM files."""
    location_ids = np.array(location_ids).astype(int)

    schema = pa.schema(
        [
            (VALUE, pa.float32()),
            (REFERENCE_TIME, pa.timestamp("ms")),
            (LOCATION_ID, pa.string()),
            (VALUE_TIME, pa.timestamp("ms")),
            (CONFIGURATION_NAME, pa.string()),
            (VARIABLE_NAME, pa.string()),
            (UNIT_NAME, pa.string()),
            (MEMBER, pa.string())
        ]
    )

    results = []
    for row in df.itertuples():
        results.append(
            file_chunk_loop(
                row,
                location_ids,
                variable_name,
                configuration,
                schema,
                ignore_missing_file,
                nwm_version,
                variable_mapper
            )
        )
    output = dask.compute(*results)

    if not any(output):
        raise FileNotFoundError("No NWM files for specified input "
                                "configuration were found in GCS!")

    output = [tbl for tbl in output if tbl is not None]
    output_table = pa.concat_tables(output)

    if process_by_z_hour:
        row = df.iloc[0]
        filename = f"{row.day}T{row.z_hour[1:3]}.parquet"
    else:
        # Use start and end dates including forecast hour
        #  for the output file name.
        filepath_list = df.filepath.sort_values().tolist()
        start_json = filepath_list[0].split("/")[-1].split(".")
        start = f"{start_json[1]}T{start_json[3][1:3]}F{start_json[6][1:]}"
        end_json = filepath_list[-1].split("/")[-1].split(".")
        end = f"{end_json[1]}T{end_json[3][1:3]}F{end_json[6][1:]}"
        filename = f"{start}_{end}.parquet"

    write_timeseries_parquet_file(
        Path(output_parquet_dir, filename),
        overwrite_output,
        output_table,
        timeseries_type
    )


def fetch_and_format_nwm_points(
    json_paths: List[str],
    location_ids: Iterable[int],
    configuration: str,
    variable_name: str,
    output_parquet_dir: str,
    process_by_z_hour: bool,
    stepsize: int,
    ignore_missing_file: bool,
    overwrite_output: bool,
    nwm_version: str,
    variable_mapper: Dict[str, Dict[str, str]],
    timeseries_type: TimeseriesTypeEnum
):
    """Fetch NWM point data and save as parquet files.

    Read in previously generated Kerchunk reference jsons,
    subset the NWM data based on provided location IDs, and format
    and save to parquet files in the TEEHR data model using Dask.

    Parameters
    ----------
    json_paths : list
        List of the single json reference filepaths.
    location_ids : Iterable[int]
        Array specifying NWM IDs of interest.
    configuration : str
        NWM forecast category.
    variable_name : str
        Name of the NWM data variable to download.
    output_parquet_dir : str
        Path to the directory for the final parquet files.
    process_by_z_hour : bool
        A boolean flag that determines the method of grouping files
        for processing.
    stepsize : int
        The number of json files to process at one time.
    ignore_missing_file : bool
        Flag specifying whether or not to fail if a missing NWM
        file is encountered
        True = skip and continue
        False = fail.
    overwrite_output : bool
        Flag specifying whether or not to overwrite output files if
        they already exist.  True = overwrite; False = fail.
    nwm_version : str
        Specified NWM version.
    """
    output_parquet_dir = Path(output_parquet_dir)
    if not output_parquet_dir.exists():
        output_parquet_dir.mkdir(parents=True)

    # Format file list into a dataframe and group by specified method
    df_refs = parse_nwm_json_paths(
        json_paths=json_paths
    )

    if process_by_z_hour:
        # Option #1. Groupby day and z_hour
        gps = df_refs.groupby(["day", "z_hour"])
        dfs = [df for _, df in gps]
    else:
        # Option #2. Chunk by some number of files
        dfs = split_dataframe(df_refs, stepsize)

    for df in dfs:
        process_chunk_of_files(
            df,
            location_ids,
            configuration,
            variable_name,
            output_parquet_dir,
            process_by_z_hour,
            ignore_missing_file,
            overwrite_output,
            nwm_version,
            variable_mapper,
            timeseries_type
        )



"""Signature models for metrics."""
from typing import List, Dict, Callable, Union

from pydantic import Field
import teehr.models.metrics.metric_attributes as tma
from teehr.models.metrics.basemodels import (
    TransformEnum,
    DeterministicBasemodel,
    BootstrapBasemodel
)
from teehr.metrics import signature_funcs as sig_funcs
from teehr.models.str_enum import StrEnum


class Count(DeterministicBasemodel):
    """Count.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "primary_count".
    func : Callable
        The function to apply to the data, by default
        :func:`signature_funcs.count`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    output_field_name: str = Field(default="count")
    transform: TransformEnum = Field(default=None)
    func: Callable = Field(default=sig_funcs.count_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value"]
    )
    attrs: Dict = Field(default=tma.COUNT_ATTRS, frozen=True)


class Minimum(DeterministicBasemodel):
    """Minimum.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "primary_minimum".
    func : Callable
        The function to apply to the data, by default
        :func:`signature_funcs.minimum`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="minimum")
    func: Callable = Field(default=sig_funcs.min_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value"]
    )
    attrs: Dict = Field(default=tma.MINIMUM_ATTRS, frozen=True)


class Maximum(DeterministicBasemodel):
    """Maximum.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "maximum".
    func : Callable
        The function to apply to the data, by default
        :func:`signature_funcs.maximum`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="maximum")
    func: Callable = Field(default=sig_funcs.max_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value"]
    )
    attrs: Dict = Field(default=tma.MAXIMUM_ATTRS, frozen=True)


class Average(DeterministicBasemodel):
    """Average.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "average".
    func : Callable
        The function to apply to the data, by default
        :func:`signature_funcs.average`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="average")
    func: Callable = Field(default=sig_funcs.avg_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value"]
    )
    attrs: Dict = Field(default=tma.AVERAGE_ATTRS, frozen=True)


class Sum(DeterministicBasemodel):
    """Sum.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "sum".
    func : Callable
        The function to apply to the data, by default
        :func:`signature_funcs.sum`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="sum")
    func: Callable = Field(default=sig_funcs.sum_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value"]
    )
    attrs: Dict = Field(default=tma.SUM_ATTRS, frozen=True)


class Variance(DeterministicBasemodel):
    """Variance.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "variance".
    func : Callable
        The function to apply to the data, by default
        :func:`signature_funcs.variance`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    bootstrap: BootstrapBasemodel = Field(default=None)
    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="variance")
    func: Callable = Field(default=sig_funcs.variance_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value"]
    )
    attrs: Dict = Field(default=tma.VARIANCE_ATTRS, frozen=True)


class MaxValueTime(DeterministicBasemodel):
    """Max Value Time.

    Parameters
    ----------
    bootstrap : DeterministicBasemodel
        The bootstrap model, by default None.
    transform : TransformEnum
        The transformation to apply to the data, by default None.
    output_field_name : str
        The output field name, by default "max_value_time".
    func : Callable
        The function to apply to the data, by default
        :func:`signature_funcs.max_value_time`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default ["primary_value"].
    attrs : Dict
        The static attributes for the metric.
    """

    transform: TransformEnum = Field(default=None)
    output_field_name: str = Field(default="max_value_time")
    func: Callable = Field(default=sig_funcs.mvt_wrapper, frozen=True)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "value_time"]
    )
    attrs: Dict = Field(default=tma.MAX_VAL_TIME_ATTRS, frozen=True)
    return_type: str = Field(default="timestamp", frozen=True)


class SignatureMetrics:
    """Define and customize signature metrics.

    Notes
    -----
    Signature metrics operate on a single field.  Available
    signature metrics are:

    - Average
    - Count
    - MaxValueTime
    - Maximum
    - Minimum
    - Sum
    - Variance
    """

    Average = Average
    Count = Count
    MaxValueTime = MaxValueTime
    Maximum = Maximum
    Minimum = Minimum
    Sum = Sum
    Variance = Variance



"""Setup an ensemble example for testing using data in the test directory."""
from pathlib import Path

import teehr
import teehr.example_data.ensemble_example_data as ensemble_example_data


def setup_hefs_example(tmpdir):
    """Set up the ensemble metrics example."""
    # Create an Evaluation object and create the directory
    ev = teehr.Evaluation(dir_path=tmpdir, create_dir=True)

    # Clone the template
    ev.clone_template()

    # Fetch the test data
    location_data_path = Path(tmpdir, "two_locations.parquet")
    ensemble_example_data.fetch_file("two_locations.parquet", location_data_path)

    crosswalk_data_path = Path(tmpdir, "location_crosswalks.parquet")
    ensemble_example_data.fetch_file("location_crosswalks.parquet", crosswalk_data_path)

    primary_timeseries_path = Path(tmpdir, "primary_timeseries.parquet")
    ensemble_example_data.fetch_file("primary_timeseries.parquet", primary_timeseries_path)

    secondary_timeseries_path = Path(tmpdir, "secondary_timeseries.parquet")
    ensemble_example_data.fetch_file("secondary_timeseries.parquet", secondary_timeseries_path)

    ev.configurations.add(
        teehr.Configuration(
            name="CNRFC_HEFS",
            type="secondary",
            description="CNRFC HEFS Extended Forecast"
        )
    )

    ev.configurations.add(
        teehr.Configuration(
            name="usgs_observations",
            type="primary",
            description="USGS Observations"
        )
    )

    # Load the data into the Evaluation
    ev.locations.load_spatial(in_path=location_data_path)
    ev.location_crosswalks.load_parquet(in_path=crosswalk_data_path)
    ev.primary_timeseries.load_parquet(in_path=primary_timeseries_path)
    ev.secondary_timeseries.load_parquet(in_path=secondary_timeseries_path)


if __name__ == "__main__":
    import shutil

    # Define the directory where the Evaluation will be created.
    test_eval_dir = Path(Path().home(), "temp", "09_ensemble_metrics")
    shutil.rmtree(test_eval_dir, ignore_errors=True)

    # Setup the example evaluation using data from the TEEHR repository.
    setup_hefs_example(tmpdir=test_eval_dir)

    # Initialize the evaluation.
    ev = teehr.Evaluation(dir_path=test_eval_dir)

    pass


"""Contains UDFs for deterministic metric calculations in Spark queries."""
import numpy as np
import numpy.typing as npt
import pandas as pd

from teehr.models.metrics.basemodels import MetricsBasemodel
from teehr.models.metrics.basemodels import TransformEnum

from typing import Callable, Optional
import logging
logger = logging.getLogger(__name__)


def _transform(
        p: pd.Series,
        s: pd.Series,
        model: MetricsBasemodel,
        t: Optional[pd.Series] = None
) -> tuple:
    """Apply timeseries transform for metrics calculations."""
    # Apply transform
    if model.transform is not None:
        match model.transform:
            case TransformEnum.log:
                logger.debug("Applying log transform")
                p = np.log(p)
                s = np.log(s)
            case TransformEnum.sqrt:
                logger.debug("Applying square root transform")
                p = np.sqrt(p)
                s = np.sqrt(s)
            case TransformEnum.square:
                logger.debug("Applying square transform")
                p = np.square(p)
                s = np.square(s)
            case TransformEnum.cube:
                logger.debug("Applying cube transform")
                p = np.power(p, 3)
                s = np.power(s, 3)
            case TransformEnum.exp:
                logger.debug("Applying exponential transform")
                p = np.exp(p)
                s = np.exp(s)
            case TransformEnum.inv:
                logger.debug("Applying inverse transform")
                p = 1.0 / p
                s = 1.0 / s
            case TransformEnum.abs:
                logger.debug("Applying absolute value transform")
                p = np.abs(p)
                s = np.abs(s)
            case _:
                raise ValueError(
                    f"Unsupported transform: {model.transform}"
                )
    else:
        logger.debug("No transform specified, using original values")

    # Remove invalid values and align series
    if t is not None:
        if isinstance(t, pd.Series):
            valid_mask = np.isfinite(p) & np.isfinite(s)
            p = p[valid_mask]
            s = s[valid_mask]
            t = t[valid_mask]
        else:
            raise TypeError(
                "t must be a pandas Series, not {type(t)}"
            )
    else:
        valid_mask = np.isfinite(p) & np.isfinite(s)
        p = p[valid_mask]
        s = s[valid_mask]

    if t is not None:
        return p, s, t
    else:
        return p, s


def _mean_error(
    y_true: npt.ArrayLike,
    y_pred: npt.ArrayLike,
    power: float = 1.0,
    root: bool = False
) -> float:
    """Mean error."""
    me = np.sum(np.abs(np.subtract(y_true, y_pred)) ** power) / len(y_true)

    # Return mean error, optionally return root mean error
    if root:
        return np.sqrt(me)
    return me


def _root_mean_squared_error(p: pd.Series, s: pd.Series) -> float:
    """Root mean squared error."""
    return _mean_error(p, s, power=2.0, root=True)


def me_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the mean_error metric function."""
    logger.debug("Building the mean_error metric function")

    def mean_error(p: pd.Series, s: pd.Series) -> float:
        """Mean Error."""
        p, s = _transform(p, s, model)
        difference = s - p
        return np.sum(difference)/len(p)

    return mean_error


def rb_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the relative_bias metric function."""
    logger.debug("Building the relative_bias metric function")

    def relative_bias(p: pd.Series, s: pd.Series) -> float:
        """Relative Bias."""
        p, s = _transform(p, s, model)
        difference = s - p
        return np.sum(difference)/np.sum(p)

    return relative_bias


def mare_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the mean_absolute_relative_error metric function."""
    logger.debug("Building the mean_absolute_relative_error metric function")

    def mean_absolute_relative_error(p: pd.Series, s: pd.Series) -> float:
        """Absolute Relative Error."""
        p, s = _transform(p, s, model)
        absolute_difference = np.abs(s - p)
        return np.sum(absolute_difference)/np.sum(p)

    return mean_absolute_relative_error


def mb_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the multiplicative_bias metric function."""
    logger.debug("Building the multiplicative_bias metric function")

    def multiplicative_bias(p: pd.Series, s: pd.Series) -> float:
        """Multiplicative Bias."""
        p, s = _transform(p, s, model)
        return np.mean(s)/np.mean(p)

    return multiplicative_bias


def pc_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the pearson_correlation metric function."""
    logger.debug("Building the pearson_correlation metric function")

    def pearson_correlation(p: pd.Series, s: pd.Series) -> float:
        """Pearson Correlation Coefficient."""
        p, s = _transform(p, s, model)
        return np.corrcoef(s, p)[0][1]

    return pearson_correlation


def r_squared_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the R-squared metric function."""
    logger.debug("Building the R-squared metric function")

    def r_squared(p: pd.Series, s: pd.Series) -> float:
        """R-squared."""
        p, s = _transform(p, s, model)
        pearson_correlation_coefficient = np.corrcoef(s, p)[0][1]
        return np.power(pearson_correlation_coefficient, 2)

    return r_squared


def mvd_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the max_value_delta metric function."""
    logger.debug("Building the max_value_delta metric function")

    def max_value_delta(p: pd.Series, s: pd.Series) -> float:
        """Max value delta."""
        p, s = _transform(p, s, model)
        return np.max(s) - np.max(p)

    return max_value_delta


def aprb_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the annual_peak_relative_bias metric function."""
    logger.debug("Building the annual_peak_relative_bias metric function")

    def annual_peak_relative_bias(
        p: pd.Series,
        s: pd.Series,
        value_time: pd.Series
    ) -> float:
        """Annual peak relative bias."""
        p, s, value_time = _transform(p, s, model, value_time)
        df = pd.DataFrame(
            {
                "primary_value": p,
                "secondary_value": s,
                "value_time": value_time
            }
        )
        primary_yearly_max_values = df.groupby(
            df.value_time.dt.year
        ).primary_value.max()
        secondary_yearly_max_values = df.groupby(
            df.value_time.dt.year
        ).secondary_value.max()
        return np.sum(
            secondary_yearly_max_values
            - primary_yearly_max_values
            ) / np.sum(primary_yearly_max_values)

    return annual_peak_relative_bias


def spearman_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the Spearman metric function."""
    logger.debug("Building the spearman_correlation metric function")

    def spearman_correlation(p: pd.Series, s: pd.Series) -> float:
        """Spearman Rank Correlation Coefficient."""
        p, s = _transform(p, s, model)

        primary_rank = p.rank()
        secondary_rank = s.rank()
        count = len(p)
        return 1 - (
            6 * np.sum(np.abs(primary_rank - secondary_rank)**2)
            / (count * (count**2 - 1))
            )

    return spearman_correlation


def nse_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the nash_sutcliffe_efficiency metric function."""
    logger.debug("Building the nash_sutcliffe_efficiency metric function")

    def nash_sutcliffe_efficiency(p: pd.Series, s: pd.Series) -> float:
        """Nash-Sutcliffe Efficiency."""
        if len(p) == 0 or len(s) == 0:
            return np.nan
        if np.sum(p) == 0 or np.sum(s) == 0:
            return np.nan

        p, s = _transform(p, s, model)

        numerator = np.sum(np.subtract(p, s) ** 2)
        denominator = np.sum(np.subtract(p, np.mean(p)) ** 2)
        if numerator == np.nan or denominator == np.nan:
            return np.nan
        if denominator == 0:
            return np.nan
        return 1.0 - numerator/denominator

    return nash_sutcliffe_efficiency


def nse_norm_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the nash_sutcliffe_efficiency_normalized metric function."""
    logger.debug(
        "Building the nash_sutcliffe_efficiency_normalized metric function"
        )

    def nash_sutcliffe_efficiency_normalized(p: pd.Series,
                                             s: pd.Series
                                             ) -> float:
        """Apply normalized Nash-Sutcliffe Efficiency."""
        if len(p) == 0 or len(s) == 0:
            return np.nan
        if np.sum(p) == 0 or np.sum(s) == 0:
            return np.nan

        p, s = _transform(p, s, model)

        numerator = np.sum(np.subtract(p, s) ** 2)
        denominator = np.sum(np.subtract(p, np.mean(p)) ** 2)
        if numerator == np.nan or denominator == np.nan:
            return np.nan
        if denominator == 0:
            return np.nan
        return 1.0 / (1.0 + numerator/denominator)

    return nash_sutcliffe_efficiency_normalized


def kge_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the kling_gupta_efficiency metric function."""
    logger.debug("Building the kling_gupta_efficiency metric function")

    def kling_gupta_efficiency(p: pd.Series,
                               s: pd.Series,
                               ) -> float:
        """Kling-Gupta Efficiency (2009)."""
        if np.std(s) == 0 or np.std(p) == 0:
            return np.nan

        p, s = _transform(p, s, model)

        # Pearson correlation coefficient
        linear_correlation = np.corrcoef(s, p)[0, 1]

        # Relative variability
        relative_variability = np.std(s) / np.std(p)

        # Relative mean
        relative_mean = np.mean(s) / np.mean(p)

        # Scaled Euclidean distance
        euclidean_distance = np.sqrt(
            (model.sr * ((linear_correlation - 1.0) ** 2.0)) +
            (model.sa * ((relative_variability - 1.0) ** 2.0)) +
            (model.sb * ((relative_mean - 1.0) ** 2.0))
        )

        # Return KGE
        return 1.0 - euclidean_distance

    return kling_gupta_efficiency


def kge_mod1_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the kling_gupta_efficiency_mod1 metric function."""
    logger.debug("Building the kling_gupta_effiency_mod1 metric function")

    def kling_gupta_efficiency_mod1(p: pd.Series, s: pd.Series) -> float:
        """Kling-Gupta Efficiency - modified 1 (2012)."""
        if np.std(s) == 0 or np.std(p) == 0:
            return np.nan

        p, s = _transform(p, s, model)

        # Pearson correlation coefficient (same as kge)
        linear_correlation = np.corrcoef(s, p)[0, 1]

        # Variability_ratio
        variability_ratio = (
            (np.std(s) / np.mean(s))
            / (np.std(p) / np.mean(p))
        )

        # Relative mean (same as kge)
        relative_mean = (np.mean(s) / np.mean(p))

        # Scaled Euclidean distance
        euclidean_distance = np.sqrt(
            (model.sr * ((linear_correlation - 1.0) ** 2.0)) +
            (model.sa * ((variability_ratio - 1.0) ** 2.0)) +
            (model.sb * ((relative_mean - 1.0) ** 2.0))
        )

        return 1.0 - euclidean_distance

    return kling_gupta_efficiency_mod1


def kge_mod2_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the kling_gupta_efficiency_mod2 metric function."""
    logger.debug("Building the kling_gupta_efficiency_mod2 metric function")

    def kling_gupta_efficiency_mod2(p: pd.Series, s: pd.Series) -> float:
        """Kling-Gupta Efficiency - modified 2 (2021)."""
        if np.std(s) == 0 or np.std(p) == 0:
            return np.nan

        p, s = _transform(p, s, model)

        # Pearson correlation coefficient (same as kge)
        linear_correlation = np.corrcoef(s, p)[0, 1]

        # Relative variability (same as kge)
        relative_variability = (np.std(s) / np.std(p))

        # bias component
        bias_component = (
            ((np.mean(s) - np.mean(p)) ** 2)
            /
            (np.std(p) ** 2)
        )

        # Scaled Euclidean distance
        euclidean_distance = np.sqrt(
            (model.sr * ((linear_correlation - 1.0) ** 2.0)) +
            (model.sa * ((relative_variability - 1.0) ** 2.0)) +
            (model.sb * bias_component)
        )

        return 1.0 - euclidean_distance

    return kling_gupta_efficiency_mod2


def mae_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the mean_absolute_error metric function."""
    logger.debug("Building the mean_absolute_error metric function")

    def mean_absolute_error(p: pd.Series, s: pd.Series) -> float:
        """Mean absolute error."""
        p, s = _transform(p, s, model)
        return _mean_error(p, s)

    return mean_absolute_error


def mse_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the mean_squared_error metric function."""
    logger.debug("Building the mean_squared_error metric function")

    def mean_squared_error(p: pd.Series, s: pd.Series) -> float:
        """Mean squared error."""
        p, s = _transform(p, s, model)
        return _mean_error(p, s, power=2.0)

    return mean_squared_error


def rmse_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the root_mean_squared_error metric function."""
    logger.debug("Building the root_mean_squared_error metric function")

    def root_mean_squared_error(p: pd.Series, s: pd.Series) -> float:
        """Root mean squared error."""
        p, s = _transform(p, s, model)
        return _mean_error(p, s, power=2.0, root=True)

    return root_mean_squared_error


def rmsdr_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the root_mean_standard_deviation_ratio metric function."""
    logger.debug(
        "Building the root_mean_standard_deviation_ratio metric function"
        )

    def root_mean_standard_deviation_ratio(p: pd.Series,
                                           s: pd.Series
                                           ) -> float:
        """Root mean standard deviation ratio."""
        p, s = _transform(p, s, model)
        rmse = _root_mean_squared_error(p, s)
        obs_std_dev = np.std(p)
        return rmse / obs_std_dev

    return root_mean_standard_deviation_ratio


# Time-based Metrics
def mvtd_wrapper(model: MetricsBasemodel) -> Callable:
    """Create the max_value_timedelta metric function."""
    logger.debug("Building the max_value_timedelta metric function")

    def max_value_timedelta(
        p: pd.Series,
        s: pd.Series,
        value_time: pd.Series
    ) -> float:
        """Max value time delta."""
        p, s, value_time = _transform(p, s, model, value_time)
        p_max_time = value_time[p.idxmax()]
        s_max_time = value_time[s.idxmax()]

        td = s_max_time - p_max_time

        return td.total_seconds()

    return max_value_timedelta



"""Convert and insert timeseries data into the dataset."""
from typing import Union
from pathlib import Path
import pandas as pd
from teehr.loading.utils import (
    merge_field_mappings,
    validate_constant_values_dict,
    read_and_convert_netcdf_to_df,
    read_and_convert_xml_to_df,
    # convert_datetime_ns_to_ms
)
import teehr.models.pandera_dataframe_schemas as schemas
from teehr.models.table_enums import TableWriteEnum

import logging

logger = logging.getLogger(__name__)


def convert_single_timeseries(
    in_filepath: Union[str, Path],
    out_filepath: Union[str, Path],
    field_mapping: dict,
    constant_field_values: dict = None,
    timeseries_type: str = None,
    **kwargs
):
    """Convert timeseries data to parquet format.

    Parameters
    ----------
    in_filepath : Union[str, Path]
        The input file path.
    out_filepath : Union[str, Path]
        The output file path.
    field_mapping : dict
        A dictionary mapping input fields to output fields.
        format: {input_field: output_field}
    constant_field_values : dict, optional
        A dictionary mapping field names to constant values.
        format: {field_name: value}
    **kwargs
        Additional keyword arguments are passed to
            pd.read_csv(), pd.read_parquet(), or xr.open_dataset().

    Steps:
    1. Read the file
    2. Rename the columns based on the field_mapping
    3. Add constant field values to dataframe,
        this may result in too many columns.
    4. Subset only the columns in the field_mapping
    5. Write the dataframe to parquet format

    """
    in_filepath = Path(in_filepath)
    out_filepath = Path(out_filepath)

    logger.info(f"Converting timeseries data from: {in_filepath}")

    if in_filepath.suffix == ".parquet":
        # read and convert parquet file
        timeseries = pd.read_parquet(in_filepath, **kwargs)
    elif in_filepath.suffix == ".csv":
        # read and convert csv file
        timeseries = pd.read_csv(in_filepath, **kwargs)
    elif in_filepath.suffix == ".nc":
        # read and convert netcdf file
        timeseries = read_and_convert_netcdf_to_df(
            in_filepath,
            field_mapping,
            **kwargs
        )
    elif in_filepath.suffix == ".xml":
        # read and convert xml file
        timeseries = read_and_convert_xml_to_df(
            in_filepath,
            field_mapping,
            **kwargs
        )
    else:
        raise ValueError("Unsupported file type.")

    timeseries.rename(columns=field_mapping, inplace=True)

    if constant_field_values:
        for field, value in constant_field_values.items():
            timeseries[field] = value

    # timeseries = timeseries[field_mapping.values()]

    if timeseries_type == "primary":
        validated_df = schemas.primary_timeseries_schema(type="pandas").validate(timeseries)
    elif timeseries_type == "secondary":
        validated_df = schemas.secondary_timeseries_schema(type="pandas").validate(timeseries)
    else:
        raise ValueError("Invalid timeseries type.")

    # validated_df = convert_datetime_ns_to_ms(validated_df)

    logger.info(f"Writing timeseries data to: {out_filepath}")
    out_filepath.parent.mkdir(parents=True, exist_ok=True)
    validated_df.to_parquet(out_filepath)


def convert_timeseries(
    in_path: Union[str, Path],
    out_path: Union[str, Path],
    field_mapping: dict = None,
    constant_field_values: dict = None,
    timeseries_type: str = None,
    pattern: str = "**/*.parquet",
    **kwargs
):
    """Convert timeseries data to parquet format.

    Parameters
    ----------
    in_path : Union[str, Path]
        The input file path.
    out_path : Union[str, Path]
        The output file path.
    field_mapping : dict, optional
        A dictionary mapping input fields to output fields.
        format: {input_field: output_field}
    constant_field_values : dict, optional
        A dictionary mapping field names to constant values.
        format: {field_name: value}
    pattern : str, optional (default: "**/*.parquet")
        The pattern to match files.
    **kwargs
        Additional keyword arguments are passed to
            pd.read_csv() or pd.read_parquet().

    Can convert CSV or Parquet files.

    """
    in_path = Path(in_path)
    out_path = Path(out_path)
    logger.info(f"Converting timeseries data: {in_path}")

    default_field_mapping = {}
    if timeseries_type == "primary":
        fields = schemas.primary_timeseries_schema(type="pandas").columns.keys()
    elif timeseries_type == "secondary":
        fields = schemas.secondary_timeseries_schema(type="pandas").columns.keys()
    else:
        raise ValueError("Invalid timeseries type.")

    for field in fields:
        if field not in default_field_mapping.values():
            default_field_mapping[field] = field

    if field_mapping:
        logger.debug("Merging user field_mapping with default field mapping.")
        field_mapping = merge_field_mappings(
            default_field_mapping,
            field_mapping
        )
    else:
        logger.debug("Using default field mapping.")
        field_mapping = default_field_mapping

    # verify constant_field_values keys are in field_mapping values
    if constant_field_values:
        validate_constant_values_dict(
            constant_field_values,
            field_mapping.values()
        )

    files_converted = 0
    if in_path.is_dir():
        # recursively convert all files in directory
        logger.info(f"Recursively converting all files in {in_path}/{pattern}")
        for in_filepath in in_path.glob(f"{pattern}"):
            relative_name = in_filepath.relative_to(in_path)
            out_filepath = Path(out_path, relative_name)
            out_filepath = out_filepath.with_suffix(".parquet")
            convert_single_timeseries(
                in_filepath,
                out_filepath,
                field_mapping,
                constant_field_values,
                timeseries_type=timeseries_type,
                **kwargs
            )
            files_converted += 1
    else:
        out_filepath = Path(out_path, in_path.name)
        out_filepath = out_filepath.with_suffix(".parquet")
        convert_single_timeseries(
            in_path,
            out_filepath,
            field_mapping,
            constant_field_values,
            timeseries_type=timeseries_type,
            **kwargs
        )
        files_converted += 1
    logger.info(f"Converted {files_converted} files.")


def validate_and_insert_timeseries(
    ev,
    in_path: Union[str, Path],
    timeseries_type: str,
    pattern: str = "**/*.parquet",
    write_mode: TableWriteEnum = "append",
):
    """Validate and insert primary timeseries data.

    Parameters
    ----------
    ev : Evaluation
        The Evaluation object.
    in_path : Union[str, Path]
        Directory path or file path to the primary timeseries data.
    timeseries_type : str
        The type of timeseries data.
        Valid values: "primary", "secondary"
    pattern : str, optional (default: "**/*.parquet")
        The pattern to match files.
    write_mode : TableWriteEnum, optional (default: "append")
        The write mode for the table. Options are "append" or "upsert".
        If "append", the Evaluation table will be appended with new data
        that does not already exist.
        If "upsert", existing data will be replaced and new data that
        does not exist will be appended.
    """
    in_path = Path(in_path)
    logger.info(f"Validating and inserting timeseries data from {in_path}")

    if timeseries_type == "primary":
        table = ev.primary_timeseries
    elif timeseries_type == "secondary":
        table = ev.secondary_timeseries
    else:
        raise ValueError("Invalid timeseries type.")

    # Read the converted files to Spark DataFrame
    df = table._read_files(in_path, pattern)

    # Validate using the _validate() method
    validated_df = table._validate(df)

    # Write to the table
    table._write_spark_df(
        validated_df,
        write_mode=write_mode
    )

    # Reload the table
    table._load_table()



"""Module for fetching and processing USGS streamflow data.

The function ``usgs_to_parquet`` fetches USGS streamflow data and saves it to
parquet files following the TEEHR data model.  The USGS tool ``dataretrieval``
is used to fetch the data from the USGS API.  Several ``chunk_by`` options are
included to allow for the data to be fetched by location_id, day, week, month,
and year (or None).

Note that the USGS API is called for each unique value in the specified
chunk_by option.  For example, if chunk_by is set to "location_id", the USGS
API will be called for each unique location_id in the provided list.

.. note::
   Care should be taken to select the appropriate chunk_by option to based on
   the number of locations and the time period of interest, to avoid excessive
   API calls and produce more efficient queries.
"""
import pandas as pd
import logging

import numpy as np
from typing import List, Union, Optional, Dict
from pathlib import Path
from datetime import datetime, timedelta
import dataretrieval.nwis as nwis
from teehr.models.fetching.utils import USGSChunkByEnum, USGSServiceEnum
from pydantic import validate_call, ConfigDict
from teehr.fetching.utils import (
    write_timeseries_parquet_file,
    get_period_start_end_times,
    create_periods_based_on_chunksize
)
from teehr.models.fetching.utils import TimeseriesTypeEnum
from teehr.fetching.const import (
    USGS_NODATA_VALUES,
    USGS_CONFIGURATION_NAME,
    VALUE,
    VALUE_TIME,
    REFERENCE_TIME,
    LOCATION_ID,
    UNIT_NAME,
    VARIABLE_NAME,
    CONFIGURATION_NAME,
    USGS_VARIABLE_MAPPER
)

DATETIME_STR_FMT = "%Y-%m-%dT%H:%M:00+0000"
DAYLIGHT_SAVINGS_PAD = timedelta(hours=2)

pd.options.mode.copy_on_write = True

logger = logging.getLogger(__name__)


def _filter_to_hourly(df: pd.DataFrame) -> pd.DataFrame:
    """Filter out data not reported on the hour."""
    logger.debug("Filtering to hourly data.")
    df.set_index(VALUE_TIME, inplace=True)
    df2 = df[
        df.index.hour.isin(range(0, 24))
        & (df.index.minute == 0)
        & (df.index.second == 0)
    ]
    df2.reset_index(level=0, allow_duplicates=True, inplace=True)
    return df2


def _filter_no_data(df: pd.DataFrame) -> pd.DataFrame:
    """Filter out no data values."""
    logger.debug("Filtering out no data values.")
    df2 = df.loc[~df[VALUE].isin(USGS_NODATA_VALUES), :]
    df2.dropna(subset=[VALUE], inplace=True)
    return df2


def _convert_to_si_units(df: pd.DataFrame, unit_name: str) -> pd.DataFrame:
    """Convert streamflow values from english to metric."""
    logger.debug("Converting to SI units.")
    df[VALUE] = df[VALUE] * 0.3048**3
    df[UNIT_NAME] = unit_name
    return df


def _format_df_column_names(
    df: pd.DataFrame,
    variable_name: str,
    unit_name: str
) -> pd.DataFrame:
    """Format dataretrieval dataframe columns to TEEHR data model."""
    logger.debug("Formatting column names.")
    df.reset_index(inplace=True)
    df.rename(
        columns={
            "site_no": LOCATION_ID,
            "00060": VALUE,
            "datetime": VALUE_TIME
        },
        inplace=True
    )
    df[LOCATION_ID] = "usgs-" + df[LOCATION_ID].astype(str)
    df[CONFIGURATION_NAME] = USGS_CONFIGURATION_NAME
    df[REFERENCE_TIME] = np.nan
    df[UNIT_NAME] = unit_name
    df[VARIABLE_NAME] = variable_name
    return df[[
        LOCATION_ID,
        REFERENCE_TIME,
        VALUE_TIME,
        VALUE,
        VARIABLE_NAME,
        UNIT_NAME,
        CONFIGURATION_NAME
    ]]


def _fetch_usgs_streamflow(
    sites: List[str],
    start_date: datetime,
    end_date: datetime,
    service: str,
    variable_mapper: Dict[str, Dict[str, str]],
    filter_to_hourly: bool = True,
    filter_no_data: bool = True,
    convert_to_si: bool = True,
) -> pd.DataFrame:
    """Fetch USGS gage data and format to TEEHR format."""
    logger.debug("Fetching USGS streamflow data from NWIS.")
    start_dt_str = start_date.strftime(DATETIME_STR_FMT)
    end_dt_str = (
        end_date
        - timedelta(minutes=1)
    ).strftime(DATETIME_STR_FMT)

    # Retrieve data --> dataretrieval
    usgs_df = nwis.get_record(
        sites=sites,
        service=service,
        start=start_dt_str,
        end=end_dt_str
    )

    variable_name = variable_mapper[VARIABLE_NAME][service]
    unit_name = variable_mapper[UNIT_NAME]["Imperial"]

    usgs_df = _format_df_column_names(usgs_df, variable_name, unit_name)

    if filter_to_hourly is True:
        usgs_df = _filter_to_hourly(usgs_df)
    if filter_no_data is True:
        usgs_df = _filter_no_data(usgs_df)
    if convert_to_si is True:
        unit_name = variable_mapper[UNIT_NAME]["SI"]
        usgs_df = _convert_to_si_units(usgs_df, unit_name)

    return usgs_df


def _format_output_filename(chunk_by: str, start_dt, end_dt) -> str:
    """Format the output filename based on min and max datetime."""
    logger.debug("Formatting output filename.")
    if chunk_by == "day":
        return f"{start_dt.strftime('%Y-%m-%d')}.parquet"
    else:
        start = start_dt.strftime('%Y-%m-%d')
        end = end_dt.strftime('%Y-%m-%d')
        return f"{start}_{end}.parquet"


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def usgs_to_parquet(
    sites: List[str],
    start_date: Union[str, datetime, pd.Timestamp],
    end_date: Union[str, datetime, pd.Timestamp],
    output_parquet_dir: Union[str, Path],
    service: USGSServiceEnum = USGSServiceEnum.iv,
    chunk_by: Union[USGSChunkByEnum, None] = None,
    filter_to_hourly: bool = True,
    filter_no_data: bool = True,
    convert_to_si: bool = True,
    overwrite_output: Optional[bool] = False,
    variable_mapper: Dict[str, Dict[str, str]] = USGS_VARIABLE_MAPPER,
    timeseries_type: TimeseriesTypeEnum = "primary"
):
    """Fetch USGS gage data and save as a Parquet file.

    All dates and times within the files and in the file names are in UTC.

    Parameters
    ----------
    sites : List[str]
        List of USGS gages sites to fetch.
        Must be string to preserve the leading 0.
    start_date : datetime
        Start time of data to fetch.
    end_date : datetime
        End time of data to fetch. Note, since startDt is inclusive for the
        USGS service, we subtract 1 minute from this time so we don't get
        overlap between consecutive calls.
    output_parquet_dir : Union[str, Path]
        Path of directory where parquet files will be saved.
    service : USGSServiceEnum
        The USGS service to use for fetching data ('iv' for hourly
        instantaneous values or 'dv' for daily mean values).
    chunk_by : Union[str, None], default = None
        How to "chunk" the fetching and storing of the data.
        Valid options = ["location_id", "day", "week", "month", "year", None].
    filter_to_hourly : bool = True
        Return only values that fall on the hour (i.e. drop 15 minute data).
    filter_no_data : bool = True
        Filter out -999 values.
    convert_to_si : bool = True
        Multiplies values by 0.3048**3 and sets `measurement_units` to `m3/s`.
    overwrite_output : bool
        Flag specifying whether or not to overwrite output files if they
        already exist.  True = overwrite; False = fail.
    varible_mapper : Dict[str]
        Dictionary mapping the USGS service to the variable name and defining
        the unit name. default: {"variable": {"iv": "streamflow_hourly_inst",
        "dv": "streamflow_daily_mean"}, "unit": {"SI": "m^3/s",
        "Imperial": "ft^3/s"}}

    Examples
    --------
    Here we fetch five days worth of USGS hourly streamflow data, to two gages,
    chunking by day.

    Import the module.

    >>> from teehr.fetching.usgs.usgs import usgs_to_parquet

    Set the input variables.

    >>> SITES=["02449838", "02450825"]
    >>> START_DATE=datetime(2023, 2, 20)
    >>> END_DATE=datetime(2023, 2, 25)
    >>> OUTPUT_PARQUET_DIR=Path(Path().home(), "temp", "usgs")
    >>> CHUNK_BY="day",
    >>> OVERWRITE_OUTPUT=True

    Fetch the data, writing to the specified output directory.

    >>> usgs_to_parquet(
    >>>     sites=SITES,
    >>>     start_date=START_DATE,
    >>>     end_date=END_DATE,
    >>>     output_parquet_dir=TEMP_DIR,
    >>>     chunk_by=CHUNK_BY,
    >>>     overwrite_output=OVERWRITE_OUTPUT
    >>> )
    """
    logger.info("Fetching USGS streamflow data.")
    start_date = pd.Timestamp(start_date)
    end_date = pd.Timestamp(end_date)

    # Check if output_parquet_dir is an existing dir
    output_dir = Path(output_parquet_dir)
    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    if chunk_by == "location_id":
        for site in sites:
            usgs_df = _fetch_usgs_streamflow(
                sites=[site],
                start_date=start_date - DAYLIGHT_SAVINGS_PAD,
                end_date=end_date + DAYLIGHT_SAVINGS_PAD,
                service=service,
                variable_mapper=variable_mapper,
                filter_to_hourly=filter_to_hourly,
                filter_no_data=filter_no_data,
                convert_to_si=convert_to_si,
            )

            usgs_df = usgs_df[(usgs_df[VALUE_TIME] >= start_date.tz_localize("UTC")) &
                              (usgs_df[VALUE_TIME] < end_date.tz_localize("UTC"))]

            if len(usgs_df) > 0:
                output_filepath = Path(
                    output_parquet_dir,
                    f"{site}.parquet"
                )
                write_timeseries_parquet_file(
                    filepath=output_filepath,
                    overwrite_output=overwrite_output,
                    data=usgs_df,
                    timeseries_type=timeseries_type
                )
        return

    # Chunk data by time
    periods = create_periods_based_on_chunksize(
        start_date=start_date,
        end_date=end_date,
        chunk_by=chunk_by
    )

    for period in periods:

        if period is not None:
            dts = get_period_start_end_times(
                period=period,
                start_date=start_date,
                end_date=end_date
            )
        else:
            dts = {"start_dt": start_date, "end_dt": end_date}

        usgs_df = _fetch_usgs_streamflow(
            sites=sites,
            start_date=dts["start_dt"] - DAYLIGHT_SAVINGS_PAD,
            end_date=dts["end_dt"] + DAYLIGHT_SAVINGS_PAD,
            service=service,
            variable_mapper=variable_mapper,
            filter_to_hourly=filter_to_hourly,
            filter_no_data=filter_no_data,
            convert_to_si=convert_to_si
        )

        usgs_df = usgs_df[(usgs_df[VALUE_TIME] >= dts["start_dt"].tz_localize("UTC")) &
                          (usgs_df[VALUE_TIME] <= dts["end_dt"].tz_localize("UTC"))]

        if len(usgs_df) > 0:

            output_filename = _format_output_filename(
                chunk_by, dts["start_dt"], dts["end_dt"]
            )

            output_filepath = Path(output_parquet_dir, output_filename)
            write_timeseries_parquet_file(
                filepath=output_filepath,
                overwrite_output=overwrite_output,
                data=usgs_df,
                timeseries_type=timeseries_type
            )


# if __name__ == "__main__":
#     # Examples
#     usgs_to_parquet(
#         sites=[
#             "02449838",
#             "02450825"
#         ],
#         start_date=datetime(2023, 2, 20),
#         end_date=datetime(2023, 2, 25),
#         output_parquet_dir=Path(Path().home(), "temp", "usgs"),
#         chunk_by="location_id",
#         overwrite_output=True
#     )

#     usgs_to_parquet(
#         sites=[
#             "02449838",
#             "02450825"
#         ],
#         start_date=datetime(2023, 2, 20),
#         end_date=datetime(2023, 2, 25),
#         output_parquet_dir=Path(Path().home(), "temp", "usgs"),
#         chunk_by="day",
#         overwrite_output=True
#     )

#     usgs_to_parquet(
#         sites=[
#             "02449838",
#             "02450825"
#         ],
#         start_date=datetime(2023, 2, 20),
#         end_date=datetime(2023, 2, 25),
#         output_parquet_dir=Path(Path().home(), "temp", "usgs"),
#         overwrite_output=True
#     )
#     pass



"""Utility function for setting up the NWM streamflow fetching example."""
from pathlib import Path

import teehr
import teehr.example_data.nwm_streamflow_example_data as fetch_nwm_streamflow_data


def setup_nwm_example(tmpdir):
    """Set up the NWM streamflow fetching example."""
    # Create an Evaluation object and create the directory
    ev = teehr.Evaluation(dir_path=tmpdir, create_dir=True)

    # Clone the template
    ev.clone_template()

    # Fetch the test data
    location_data_path = Path(tmpdir, "usgs_at_radford_location.parquet")
    fetch_nwm_streamflow_data.fetch_file("usgs_at_radford_location.parquet", location_data_path)

    crosswalk_data_path = Path(tmpdir, "location_crosswalks.parquet")
    fetch_nwm_streamflow_data.fetch_file("location_crosswalks.parquet", crosswalk_data_path)

    primary_timeseries_path = Path(tmpdir, "usgs_timeseries.parquet")
    fetch_nwm_streamflow_data.fetch_file("usgs_timeseries.parquet", primary_timeseries_path)

    secondary_timeseries_path = Path(tmpdir, "nwm_forecasts.parquet")
    fetch_nwm_streamflow_data.fetch_file("nwm_forecasts.parquet", secondary_timeseries_path)

    ev.configurations.add(
        teehr.Configuration(
            name="nwm30_medium_range",
            type="secondary",
            description="NWM Medium Range Streamflow Ensemble"
        )
    )

    ev.configurations.add(
        teehr.Configuration(
            name="usgs_observations",
            type="primary",
            description="USGS Observations"
        )
    )

    # Load the data into the Evaluation
    ev.locations.load_spatial(in_path=location_data_path)
    ev.location_crosswalks.load_parquet(in_path=crosswalk_data_path)
    ev.primary_timeseries.load_parquet(in_path=primary_timeseries_path)
    ev.secondary_timeseries.load_parquet(in_path=secondary_timeseries_path)


if __name__ == "__main__":
    import shutil

    # Define the directory where the Evaluation will be created.
    test_eval_dir = Path(Path().home(), "temp", "10_fetch_nwm_data")
    shutil.rmtree(test_eval_dir, ignore_errors=True)

    # Setup the example evaluation using data from the TEEHR repository.
    setup_nwm_example(tmpdir=test_eval_dir)

    # Initialize the evaluation.
    ev = teehr.Evaluation(dir_path=test_eval_dir)

    pass


"""Example script for setting up a new evaluation.

This script demonstrates how to set up a new evaluation using the Evaluation
class.


"""
from teehr import Evaluation
from pathlib import Path
from teehr.models.pydantic_table_models import (
    Attribute,
)

# Set a path to the directory where the evaluation will be created
TEST_STUDY_DIR = Path(Path().home(), "temp", "test_study")

# Set a path to the directory where the test data is stored
TEST_DATA_DIR = Path("/home/sam/git_local/teehr/tests/data/v0_3_test_study")
GEOJSON_GAGES_FILEPATH = Path(TEST_DATA_DIR, "geo", "gages.geojson")
PRIMARY_TIMESERIES_FILEPATH = Path(
    TEST_DATA_DIR, "timeseries", "test_short_obs.parquet"
)
CROSSWALK_FILEPATH = Path(TEST_DATA_DIR, "geo", "crosswalk.csv")
SECONDARY_TIMESERIES_FILEPATH = Path(
    TEST_DATA_DIR, "timeseries", "test_short_fcast.parquet"
)
GEO_FILEPATH = Path(TEST_DATA_DIR, "geo")

# Create an Evaluation object
ev = Evaluation(dir_path=TEST_STUDY_DIR)

# Enable logging
ev.enable_logging()

# Clone the template
ev.clone_template()

# Load the location data
ev.locations.load_spatial(in_path=GEOJSON_GAGES_FILEPATH)


# Load the timeseries data and map over the fields and set constants
ev.primary_timeseries.load_parquet(
    in_path=PRIMARY_TIMESERIES_FILEPATH,
    field_mapping={
        "reference_time": "reference_time",
        "value_time": "value_time",
        "configuration": "configuration_name",
        "measurement_unit": "unit_name",
        "variable_name": "variable_name",
        "value": "value",
        "location_id": "location_id"
    },
    constant_field_values={
        "unit_name": "m^3/s",
        "variable_name": "streamflow_hourly_inst",
        "configuration_name": "usgs_observations"
    }
)

# Load the crosswalk data
ev.location_crosswalks.load_csv(
    in_path=CROSSWALK_FILEPATH
)

# Load the secondary timeseries data and map over the fields and set constants
ev.secondary_timeseries.load_parquet(
    in_path=SECONDARY_TIMESERIES_FILEPATH,
    field_mapping={
        "reference_time": "reference_time",
        "value_time": "value_time",
        "configuration": "configuration_name",
        "measurement_unit": "unit_name",
        "variable_name": "variable_name",
        "value": "value",
        "location_id": "location_id"
    },
    constant_field_values={
        "unit_name": "m^3/s",
        "variable_name": "streamflow_hourly_inst",
        "configuration_name": "nwm30_retrospective"
    }
)

# Add some attributes
ev.attributes.add(
    [
        Attribute(
            name="drainage_area",
            type="continuous",
            description="Drainage area in square kilometers"
        ),
        Attribute(
            name="ecoregion",
            type="categorical",
            description="Ecoregion"
        ),
        Attribute(
            name="year_2_discharge",
            type="continuous",
            description="2-yr discharge in cubic meters per second"
        ),
    ]
)

# Load the location attribute data
ev.location_attributes.load_parquet(
    in_path=GEO_FILEPATH,
    field_mapping={"attribute_value": "value"},
    pattern="test_attr_*.parquet",
)

# Create the joined timeseries
ev.joined_timeseries.create(execute_scripts=True)



"""Contains functions for bootstrap calculations for use in Spark queries."""
from typing import Dict, Callable
import logging

import pandas as pd
import numpy as np
from arch.bootstrap import (
    StationaryBootstrap,
    CircularBlockBootstrap,
)

from teehr.models.metrics.basemodels import MetricsBasemodel
from teehr.metrics.gumboot_bootstrap import GumbootBootstrap

logger = logging.getLogger(__name__)


def _calculate_quantiles(
    output_field_name: str,
    results: np.ndarray,
    quantiles: list
) -> Dict:
    """Calculate quantile values of the bootstrap results."""
    values = np.quantile(results, quantiles)
    quantiles = [f"{output_field_name}_{str(i)}" for i in quantiles]
    d = dict(zip(quantiles, values))
    return d


def create_circularblock_func(model: MetricsBasemodel) -> Callable:
    """Create the CircularBlock bootstrap function."""
    logger.debug("Building the Circular Block bootstrap func.")

    def bootstrap_func(p: pd.Series, s: pd.Series) -> Dict:
        """Bootstrap function."""
        bs = CircularBlockBootstrap(
            model.bootstrap.block_size,
            p,
            s,
            seed=model.bootstrap.seed,
            random_state=model.bootstrap.random_state
        )

        results = bs.apply(
            model.func(model),
            model.bootstrap.reps
        )

        if model.bootstrap.quantiles is not None:
            return _calculate_quantiles(
                model.output_field_name,
                results,
                model.bootstrap.quantiles,
            )
        else:
            return results.ravel()
    return bootstrap_func


def create_gumboot_func(model: MetricsBasemodel) -> Callable:
    """Create the Gumboot bootstrap function."""
    logger.debug("Building the Gumboot bootstrap func.")

    def bootstrap_func(p: pd.Series, s: pd.Series, vt: pd.Series) -> Dict:
        """Bootstrap function."""
        bs = GumbootBootstrap(
            p,
            s,
            value_time=vt,
            seed=model.bootstrap.seed,
            water_year_month=model.bootstrap.water_year_month,
            boot_year_file=model.bootstrap.boot_year_file
        )

        results = bs.apply(
            model.func(model),
            model.bootstrap.reps
        )

        if model.bootstrap.quantiles is not None:
            return _calculate_quantiles(
                model.output_field_name,
                results,
                model.bootstrap.quantiles,
            )
        else:
            return results.ravel()
    return bootstrap_func


def create_stationary_func(model: MetricsBasemodel) -> Callable:
    """Create the Stationary bootstrap function."""
    logger.debug("Building the Stationary bootstrap func.")

    def bootstrap_func(p: pd.Series, s: pd.Series) -> Dict:
        """Bootstrap function."""
        bs = StationaryBootstrap(
            model.bootstrap.block_size,
            p,
            s,
            seed=model.bootstrap.seed,
            random_state=model.bootstrap.random_state
        )

        results = bs.apply(
            model.func(model),
            model.bootstrap.reps
        )

        if model.bootstrap.quantiles is not None:
            return _calculate_quantiles(
                model.output_field_name,
                results,
                model.bootstrap.quantiles,
            )
        else:
            return results.ravel()
    return bootstrap_func



"""Module describing NWM v2.2 grid configuration variables."""
from teehr.models.str_enum import StrEnum

from typing import Optional

from pydantic import BaseModel


# GRID DATA VARIABLES ENUMS
class LandAssimVariablesEnum(StrEnum):
    """LandAssimVariablesEnum."""
    ACCET = "ACCET"
    ACSNOM = "ACSNOM"
    EDIR = "EDIR"
    FSNO = "FSNO"
    ISNOW = "ISNOW"
    QRAIN = "QRAIN"
    QSNOW = "QSNOW"
    SNEQV = "SNEQV"
    SNLIQ = "SNLIQ"
    SNOWH = "SNOWH"
    SNOWT_AVG = "SNOWT_AVG"
    SOILICE = "SOILICE"
    SOILSAT_TOP = "SOILSAT_TOP"
    SOIL_M = "SOIL_M"
    SOIL_T = "SOIL_T"


class LandShortVariablesEnum(StrEnum):
    """LandShortVariablesEnum."""
    ACCET = "ACCET"
    SNOWT_AVG = "SNOWT_AVG"
    SOILSAT_TOP = "SOILSAT_TOP"
    FSNO = "FSNO"
    SNOWH = "SNOWH"
    SNEQV = "SNEQV"


class LandMediumVariablesEnum(StrEnum):
    """LandMediumVariablesEnum."""
    FSA = "FSA"
    FIRA = "FIRA"
    GRDFLX = "GRDFLX"
    HFX = "HFX"
    LH = "LH"
    UGDRNOFF = "UGDRNOFF"
    ACCECAN = "ACCECAN"
    ACCEDIR = "ACCEDIR"
    ACCETRAN = "ACCETRAN"
    TRAD = "TRAD"
    SNLIQ = "SNLIQ"
    SOIL_T = "SOIL_T"
    SOIL_M = "SOIL_M"
    SNOWH = "SNOWH"
    SNEQV = "SNEQV"
    ISNOW = "ISNOW"
    FSNO = "FSNO"
    ACSNOM = "ACSNOM"
    ACCET = "ACCET"
    CANWAT = "CANWAT"
    SOILICE = "SOILICE"
    SOILSAT_TOP = "SOILSAT_TOP"
    SNOWT_AVG = "SNOWT_AVG"


class LandLongVariablesEnum(StrEnum):
    """LandLongVariablesEnum."""
    UGDRNOFF = "UGDRNOFF"
    SFCRNOFF = "SFCRNOFF"
    SNEQV = "SNEQV"
    ACSNOM = "ACSNOM"
    ACCET = "ACCET"
    CANWAT = "CANWAT"
    SOILSAT_TOP = "SOILSAT_TOP"
    SOILSAT = "SOILSAT"


class ForcingVariablesEnum(StrEnum):
    """ForcingVariablesEnum."""
    U2D = "U2D"
    V2D = "V2D"
    T2D = "T2D"
    Q2D = "Q2D"
    LWDOWN = "LWDOWN"
    SWDOWN = "SWDOWN"
    RAINRATE = "RAINRATE"
    PSFC = "PSFC"


# OUTPUT TYPE ENUMS
class ShortAndAnalysisOutputEnum(StrEnum):
    """ShortAndAnalysisOutputEnum."""
    land = "land"


class MediumOutputEnum(StrEnum):
    """MediumOutputEnum."""
    land_1 = "land_1"
    land_2 = "land_2"
    land_3 = "land_3"
    land_4 = "land_4"
    land_5 = "land_5"
    land_6 = "land_6"
    land_7 = "land_7"


class LongOutputEnum(StrEnum):
    """LongOutputEnum."""
    land_1 = "land_1"
    land_2 = "land_2"
    land_3 = "land_3"
    land_4 = "land_4"


class ForcingOutputEnum(StrEnum):
    """ForcingOutputEnum."""
    forcing = "forcing"


# OUTPUT TYPE MODELS
class Analysis(BaseModel):
    """Analysis."""
    output_type: ShortAndAnalysisOutputEnum
    land: Optional[LandAssimVariablesEnum] = None


class ShortRange(BaseModel):
    """ShortRange."""
    output_type: ShortAndAnalysisOutputEnum
    land: Optional[LandShortVariablesEnum] = None


class MediumRange(BaseModel):
    """MediumRange."""
    output_type: MediumOutputEnum
    land_1: Optional[LandMediumVariablesEnum] = None
    land_2: Optional[LandMediumVariablesEnum] = None
    land_3: Optional[LandMediumVariablesEnum] = None
    land_4: Optional[LandMediumVariablesEnum] = None
    land_5: Optional[LandMediumVariablesEnum] = None
    land_6: Optional[LandMediumVariablesEnum] = None
    land_7: Optional[LandMediumVariablesEnum] = None


class LongRange(BaseModel):
    """LongRange."""
    output_type: LongOutputEnum
    land_1: Optional[LandLongVariablesEnum] = None
    land_2: Optional[LandLongVariablesEnum] = None
    land_3: Optional[LandLongVariablesEnum] = None
    land_4: Optional[LandLongVariablesEnum] = None


class Forcing(BaseModel):
    """Forcing."""
    output_type: ForcingOutputEnum
    forcing: Optional[ForcingVariablesEnum] = None


# CONFIGURATIONS ENUM
class ConfigurationsEnum(StrEnum):
    """ConfigurationsEnum."""
    analysis_assim = "analysis_assim"
    analysis_assim_no_da = "analysis_assim_no_da"
    analysis_assim_extend = "analysis_assim_extend"
    analysis_assim_extend_no_da = "analysis_assim_extend_no_da"
    analysis_assim_long = "analysis_assim_long"
    analysis_assim_long_no_da = "analysis_assim_long_no_da"
    analysis_assim_hawaii = "analysis_assim_hawaii"
    analysis_assim_hawaii_no_da = "analysis_assim_hawaii_no_da"
    analysis_assim_puertorico = "analysis_assim_puertorico"
    analysis_assim_puertorico_no_da = "analysis_assim_puertorico_no_da"
    short_range = "short_range"
    short_range_hawaii = "short_range_hawaii"
    short_range_puertorico = "short_range_puertorico"
    short_range_hawaii_no_da = "short_range_hawaii_no_da"
    short_range_puertorico_no_da = "short_range_puertorico_no_da"
    medium_range_mem1 = "medium_range_mem1"
    medium_range_mem2 = "medium_range_mem2"
    medium_range_mem3 = "medium_range_mem3"
    medium_range_mem4 = "medium_range_mem4"
    medium_range_mem5 = "medium_range_mem5"
    medium_range_mem6 = "medium_range_mem6"
    medium_range_mem7 = "medium_range_mem7"
    long_range_mem1 = "long_range_mem1"
    long_range_mem2 = "long_range_mem2"
    long_range_mem3 = "long_range_mem3"
    long_range_mem4 = "long_range_mem4"
    forcing_medium_range = "forcing_medium_range"
    forcing_short_range = "forcing_short_range"
    forcing_short_range_hawaii = "forcing_short_range_hawaii"
    forcing_short_range_puertorico = "forcing_short_range_puertorico"
    forcing_analysis_assim = "forcing_analysis_assim"
    forcing_analysis_assim_extend = "forcing_analysis_assim_extend"
    forcing_analysis_assim_hawaii = "forcing_analysis_assim_hawaii"
    forcing_analysis_assim_puertorico = "forcing_analysis_assim_puertorico"


# CONFIGURATION MODEL
class GridConfigurationModel(BaseModel):
    """NWM v2.2 GridConfigurationModel."""
    configuration: ConfigurationsEnum
    analysis_assim: Optional[Analysis] = None
    analysis_assim_no_da: Optional[Analysis] = None
    analysis_assim_extend: Optional[Analysis] = None
    analysis_assim_extend_no_da: Optional[Analysis] = None
    analysis_assim_long: Optional[Analysis] = None
    analysis_assim_long_no_da: Optional[Analysis] = None
    analysis_assim_hawaii: Optional[Analysis] = None
    analysis_assim_hawaii_no_da: Optional[Analysis] = None
    analysis_assim_puertorico: Optional[Analysis] = None
    analysis_assim_puertorico_no_da: Optional[Analysis] = None
    short_range: Optional[ShortRange] = None
    short_range_hawaii: Optional[ShortRange] = None
    short_range_puertorico: Optional[ShortRange] = None
    short_range_hawaii_no_da: Optional[ShortRange] = None
    short_range_puertorico_no_da: Optional[ShortRange] = None
    medium_range_mem1: Optional[MediumRange] = None
    medium_range_mem2: Optional[MediumRange] = None
    medium_range_mem3: Optional[MediumRange] = None
    medium_range_mem4: Optional[MediumRange] = None
    medium_range_mem5: Optional[MediumRange] = None
    medium_range_mem6: Optional[MediumRange] = None
    medium_range_mem7: Optional[MediumRange] = None
    medium_range_no_da: Optional[MediumRange] = None
    long_range_mem1: Optional[LongRange] = None
    long_range_mem2: Optional[LongRange] = None
    long_range_mem3: Optional[LongRange] = None
    long_range_mem4: Optional[LongRange] = None
    forcing_medium_range: Optional[Forcing] = None
    forcing_short_range: Optional[Forcing] = None
    forcing_short_range_hawaii: Optional[Forcing] = None
    forcing_short_range_puertorico: Optional[Forcing] = None
    forcing_analysis_assim: Optional[Forcing] = None
    forcing_analysis_assim_extend: Optional[Forcing] = None
    forcing_analysis_assim_hawaii: Optional[Forcing] = None
    forcing_analysis_assim_puertorico: Optional[Forcing] = None


if __name__ == "__main__":
    # So for example:
    configuration = "forcing_medium_range"
    output_type = "forcing"
    variable_name = "RAINRATE"

    # Assemble input parameters
    vars = {
        "configuration": configuration,
        configuration: {
            "output_type": output_type,
            output_type: variable_name,
        },
    }

    # import sys
    # Check input parameters
    # try:
    cm = GridConfigurationModel.model_validate(vars)
    # except ValidationError as e:
    #     print(e.errors()[0]["msg"])
    #     sys.exit()

    config = cm.configuration.name
    forecast_obj = getattr(cm, config)
    out_type = forecast_obj.output_type.name
    var_name = getattr(forecast_obj, out_type).name

    pass



from teehr.evaluation.tables.domain_table import DomainTable
from teehr.models.filters import ConfigurationFilter
from teehr.models.table_enums import ConfigurationFields
from teehr.models.pydantic_table_models import Configuration
import teehr.models.pandera_dataframe_schemas as schemas
from typing import List, Union
from teehr.utils.utils import to_path_or_s3path


class ConfigurationTable(DomainTable):
    """Access methods to configurations table."""

    def __init__(self, ev):
        """Initialize class."""
        super().__init__(ev)
        self.name = "configurations"
        self.dir = to_path_or_s3path(ev.dataset_dir, self.name)
        self.filter_model = ConfigurationFilter
        self.schema_func = schemas.configuration_schema
        self.unique_column_set = ["name"]

    def field_enum(self) -> ConfigurationFields:
        """Get the configuration fields enum."""
        fields = self._get_schema("pandas").columns.keys()
        return ConfigurationFields(
            "ConfigurationFields",
            {field: field for field in fields}
        )

    def add(
        self,
        configuration: Union[Configuration, List[Configuration]]
    ):
        """Add a configuration domain to the evaluation.

        Parameters
        ----------
        configuration : Union[Configuration, List[Configuration]]
            The configuration domain to add.

        Example
        -------
        >>> from teehr.models.domain_tables import Configuration
        >>> configuration = Configuration(
        >>>     name="usgs_observations",
        >>>     type="primary",
        >>>     description="USGS observations",
        >>> )
        >>> ev.load.add_configuration(configuration)

        """
        self._add(configuration)


"""Functions for formatting metrics for querying."""
from typing import List
import logging

import pandas as pd
from pyspark.sql import GroupedData
from pyspark.sql.functions import pandas_udf

from teehr.models.metrics.basemodels import MetricsBasemodel
from teehr.models.metrics.basemodels import MetricCategories as mc
from teehr.querying.utils import validate_fields_exist, parse_fields_to_list

logger = logging.getLogger(__name__)


def apply_aggregation_metrics(
    gp: GroupedData,
    include_metrics: List[MetricsBasemodel] = None
) -> pd.DataFrame:
    """Apply metrics to a PySpark DataFrame."""
    if not isinstance(include_metrics, List):
        include_metrics = [include_metrics]

    func_list = []
    for model in include_metrics:

        input_field_names = parse_fields_to_list(model.input_field_names)
        validate_fields_exist(gp._df.columns, input_field_names)

        alias = model.output_field_name

        if "bootstrap" in model.model_dump() and model.bootstrap is not None:
            logger.debug(
                f"Applying metric: {alias} with {model.bootstrap.name}"
                " bootstrapping"
            )

            func_pd = pandas_udf(
                model.bootstrap.func(model),
                model.bootstrap.return_type
            )
            if (model.bootstrap.include_value_time) and \
               ("value_time" not in input_field_names):
                input_field_names.append("value_time")
        else:
            logger.debug(f"Applying metric: {alias}")
            func_pd = pandas_udf(model.func(model), model.return_type)

        func_list.append(
            func_pd(*input_field_names).alias(alias)
        )

    sdf = gp.agg(*func_list)

    # Note: Test exploding multiple columns
    for model in include_metrics:
        if model.unpack_results:
            sdf = model.unpack_function(sdf, model.output_field_name)

    return sdf



import pandera.pyspark as ps
import pandera as pa
import pyspark.sql.types as T
import pandas as pd
from typing import List, Union

# Domains
# PySpark Pandera Schemas

def desc_no_commas(pyspark_obj) -> bool:
    """Ensure description column does not contain commas."""
    return pyspark_obj.filter(pyspark_obj["description"].contains(",")).count() == 0


def ln_no_commas(pyspark_obj) -> bool:
    """Ensure description column does not contain commas."""
    return pyspark_obj.filter(pyspark_obj["long_name"].contains(",")).count() == 0


def valid_name(pyspark_obj) -> bool:
    """Ensure name column matches ^[a-zA-Z0-9_]+$."""
    return pyspark_obj.filter(pyspark_obj["name"].rlike(r"^[a-zA-Z0-9_]+$")).count() == pyspark_obj.count()


def valid_unit_name(pyspark_obj) -> bool:
    """Ensure name column matches ^[a-zA-Z0-9_^/]+$."""
    return pyspark_obj.filter(pyspark_obj["name"].rlike(r"^[a-zA-Z0-9_^/]+$")).count() == pyspark_obj.count()

def format_datetime64(s: pd.Series) -> pd.Series:
    # Convert to UTC.
    # if s.dt.tz is not None:
    #     s = s.dt.tz_convert("UTC")
    # Drop timezone information.
    s = s.dt.tz_localize(None)
    return s.astype("datetime64[ms]")

def configuration_schema(type: str = "pyspark") -> ps.DataFrameSchema:
    if type == "pandas":
        return pa.DataFrameSchema(
            columns={
                "name": pa.Column(
                    pa.String,
                    unique=True,
                    checks=[
                        pa.Check.str_matches(r"^[a-zA-Z0-9_]+$")
                    ],
                    regex=r"^[a-zA-Z0-9_]+$"
                ),
                "type": pa.Column(
                    pa.String,
                    checks=pa.Check.isin(
                        ["primary", "secondary"]
                    )
                ),
                "description": pa.Column(
                    pa.String,
                    checks=pa.Check(lambda s: not s.str.contains(",").any())
                ),
            },
            strict="filter"
        )
    if type == "pyspark":
        return ps.DataFrameSchema(
            columns={
                "name": ps.Column(
                    T.StringType(),
                    nullable=False,
                    checks=ps.Check(valid_name, error="`name` column contains special characters")
                ),
                "type": ps.Column(
                    T.StringType(),
                    nullable=False,
                    checks=ps.Check.isin(["primary", "secondary"])
                ),
                "description": ps.Column(
                    T.StringType(),
                    nullable=False,
                    checks=ps.Check(desc_no_commas, error="`description` column contains commas")
                ),
            },
            strict=True,
            unique=["name"],
            coerce=True
        )

def unit_schema(type: str = "pyspark") -> ps.DataFrameSchema:
    if type == "pandas":
        return pa.DataFrameSchema(
            columns={
                "name": pa.Column(
                    pa.String,
                    unique=True,
                    checks=[
                        pa.Check.str_matches(r"^[a-zA-Z0-9_^/]+$")
                    ]
                ),
                "long_name": pa.Column(
                    pa.String,
                    checks=pa.Check(lambda s: not s.str.contains(",").any())
                ),
            },
            strict="filter"
        )
    if type == "pyspark":
        return ps.DataFrameSchema(
            columns={
                "name": ps.Column(
                    T.StringType(),
                    nullable=False,
                    checks=ps.Check(valid_unit_name, error="`name` column contains special characters")
                ),
                "long_name": ps.Column(
                    T.StringType(),
                    nullable=False,
                    checks=ps.Check(ln_no_commas, error="`long_name` column contains commas")
                ),
            },
            strict=True,
            unique=["name"],
            coerce=True
        )

def variable_schema(type: str = "pyspark") -> ps.DataFrameSchema:
    if type == "pandas":
        return pa.DataFrameSchema(
            columns={
                "name": pa.Column(
                    pa.String,
                    unique=True,
                    checks=[
                        pa.Check.str_matches(r"^[a-zA-Z0-9_]+$")
                    ],
                    regex=r"^[a-zA-Z0-9_]+$"
                ),
                "long_name": pa.Column(
                    pa.String,
                    checks=pa.Check(lambda s: not s.str.contains(",").any())
                ),
            },
            strict="filter"
        )
    if type == "pyspark":
        return ps.DataFrameSchema(
            columns={
                "name": ps.Column(
                    T.StringType(),
                    nullable=False,
                    checks=ps.Check(valid_name, error="`name` column contains special characters")
                ),
                "long_name": ps.Column(
                    T.StringType(),
                    nullable=False,
                    checks=ps.Check(ln_no_commas, error="`long_name` column contains commas")
                ),
            },
            strict=True,
            unique=["name"],
            coerce=True
        )

def attribute_schema(type: str = "pyspark") -> ps.DataFrameSchema:
    if type == "pandas":
        return pa.DataFrameSchema(
            columns={
                "name": pa.Column(
                    pa.String,
                    unique=True,
                    checks=[
                        pa.Check.str_matches(r"^[a-zA-Z0-9_]+$")
                    ],
                ),
                "type": pa.Column(
                    pa.String,
                    checks=pa.Check.isin(
                        ["categorical", "continuous"]
                    )
                ),
                "description": pa.Column(
                    pa.String,
                    checks=pa.Check(lambda s: not s.str.contains(",").any())
                ),
            },
            strict="filter"
        )
    if type == "pyspark":
        return ps.DataFrameSchema(
            columns={
                "name": ps.Column(
                    T.StringType(),
                    nullable=False,
                    checks=ps.Check(valid_name, error="`name` column contains special characters")
                ),
                "type": ps.Column(
                    T.StringType(),
                    nullable=False,
                    checks=ps.Check.isin(["categorical", "continuous"])
                ),
                "description": ps.Column(
                    T.StringType(),
                    nullable=False,
                    checks=ps.Check(desc_no_commas, error="`description` column contains commas")
                ),
            },
            strict=True,
            unique=["name"],
            coerce=True
        )

# Locations
# PySpark Pandera Models
def locations_schema(type: str = "pyspark") -> ps.DataFrameSchema:
    if type == "pandas":
        return pa.DataFrameSchema(
            columns={
                "id": pa.Column(str, coerce=True),
                "name": pa.Column(str, coerce=True),
                "geometry": pa.Column("geometry")
            },
            strict="filter"
        )
    if type == "pyspark":
        return ps.DataFrameSchema(
            columns={
                "id": ps.Column(T.StringType(), nullable=False),
                "name": ps.Column(T.StringType(), nullable=False),
                "geometry": ps.Column(T.BinaryType(), nullable=False),
            },
            strict=True,
            unique=["id"],
            coerce=True
        )

def location_attributes_schema(
        location_ids: List[str] = [None],
        attr_names: List[str] = [None],
        type: str = "pyspark",
) -> ps.DataFrameSchema:
    if type == "pandas":
        return pa.DataFrameSchema(
            columns={
                "location_id": pa.Column(
                    pa.String,
                    # pa.Check.isin(location_ids),
                    coerce=True
                ),
                "attribute_name": pa.Column(
                    pa.String,
                    # pa.Check.isin(attr_names),
                    coerce=True
                ),
                "value": pa.Column(
                    pa.String,
                    coerce=True
                )
            },
            strict="filter"
        )
    if type == "pyspark":
        return ps.DataFrameSchema(
            columns={
                "location_id": ps.Column(
                    T.StringType(),
                    ps.Check.isin(location_ids),
                    nullable=False
                ),
                "attribute_name": ps.Column(
                    T.StringType(),
                    ps.Check.isin(attr_names),
                    nullable=False
                ),
                "value": ps.Column(
                    T.StringType(),
                    nullable=False,
                )
            },
            strict=True,
            coerce=True,
        )

def location_crosswalks_schema(
        location_ids: List[str] = [None],
        type: str = "pyspark",
) -> ps.DataFrameSchema:
    if type == "pandas":
        return pa.DataFrameSchema(
            columns={
                "primary_location_id": pa.Column(
                    pa.String,
                    # pa.Check.isin(location_ids),
                    coerce=True
                ),
                "secondary_location_id": pa.Column(
                    pa.String,
                    coerce=True
                )
            },
            strict="filter"
        )
    if type == "pyspark":
        return ps.DataFrameSchema(
            columns={
                "primary_location_id": ps.Column(
                    T.StringType(),
                    ps.Check.isin(location_ids),
                    nullable=False,
                ),
                "secondary_location_id": ps.Column(
                    T.StringType(),
                    nullable=False,
                )
            },
            strict=True,
            coerce=True,
        )


def weights_file_schema() -> pa.DataFrameSchema:
    """Return the schema for a weights file."""
    return pa.DataFrameSchema(
        columns={
            "row": pa.Column(
                pa.Int32,
                nullable=False,
            ),
            "col": pa.Column(
                pa.Int32,
                nullable=False,
            ),
            "weight": pa.Column(
                pa.Float32,
                nullable=False,
            ),
            "location_id": pa.Column(
                pa.String,
                nullable=False,
            )
        },
        strict="filter",
        coerce=True
    )


# Timeseries
pandas_value_type = pa.Float32()
pyspark_value_type = T.FloatType()

# PySpark Pandera Models
def primary_timeseries_schema(
        location_ids: List[str] = [None],
        variable_names: List[str] = [None],
        configuration_names: List[str] = [None],
        unit_names: List[str] = [None],
        type: str = "pyspark",
) -> ps.DataFrameSchema:
    if type == "pandas":
        return pa.DataFrameSchema(
            columns={
                "reference_time": pa.Column(
                    pa.DateTime,
                    parsers=pa.Parser(format_datetime64),
                    nullable=True,
                    coerce=True,
                ),
                "value_time": pa.Column(
                    pa.DateTime,
                    parsers=pa.Parser(format_datetime64),
                    nullable=False,
                    coerce=True,
                ),
                "value": pa.Column(
                    pandas_value_type,
                    nullable=False,
                    coerce=True,
                ),
                "variable_name": pa.Column(
                    pa.String,
                    # checks=pa.Check.isin(variable_names),
                    nullable=False
                ),
                "configuration_name": pa.Column(
                    pa.String,
                    # checks=pa.Check.isin(configuration_names),
                    nullable=False
                ),
                "unit_name": pa.Column(
                    pa.String,
                    # checks=pa.Check.isin(unit_names),
                    nullable=False
                ),
                "location_id": pa.Column(
                    pa.String,
                    # checks=pa.Check.isin(location_ids),
                    nullable=False
                )
            },
            strict="filter",
            coerce=True
        )
    if type == "pyspark":
        return ps.DataFrameSchema(
            columns={
                "reference_time": ps.Column(
                    T.TimestampNTZType(),
                    nullable=True
                ),
                "value_time": ps.Column(
                    T.TimestampNTZType(),
                    nullable=False
                ),
                "value": ps.Column(
                    pyspark_value_type,
                    nullable=False,
                    coerce=True,
                ),
                "variable_name": ps.Column(
                    T.StringType(),
                    ps.Check.isin(variable_names),
                    nullable=False
                ),
                "configuration_name": ps.Column(
                    T.StringType(),
                    ps.Check.isin(configuration_names),
                    nullable=False
                ),
                "unit_name": ps.Column(
                    T.StringType(),
                    ps.Check.isin(unit_names),
                    nullable=False
                ),
                "location_id": ps.Column(
                    T.StringType(),
                    ps.Check.isin(location_ids),
                    nullable=False
                )
            },
            strict=True,
            coerce=True,
        )


def secondary_timeseries_schema(
        location_ids: List[str] = [None],
        variable_names: List[str] = [None],
        configuration_names: List[str] = [None],
        unit_names: List[str] = [None],
        type: str = "pyspark",
) -> ps.DataFrameSchema:
    if type == "pandas":
        return pa.DataFrameSchema(
            columns={
                "reference_time": pa.Column(
                    pa.DateTime,
                    parsers=pa.Parser(format_datetime64),
                    nullable=True
                ),
                "value_time": pa.Column(
                    pa.DateTime,
                    parsers=pa.Parser(format_datetime64),
                    nullable=False
                ),
                "value": pa.Column(
                    pandas_value_type,
                    nullable=False,
                    coerce=True,
                ),
                "variable_name": pa.Column(
                    pa.String,
                    # checks=pa.Check.isin(variable_names),
                    nullable=False
                ),
                "configuration_name": pa.Column(
                    pa.String,
                    # checks=pa.Check.isin(configuration_names),
                    nullable=False
                ),
                "unit_name": pa.Column(
                    pa.String,
                    # checks=pa.Check.isin(unit_names),
                    nullable=False
                ),
                "location_id": pa.Column(
                    pa.String,
                    # checks=pa.Check.isin(location_ids),
                    nullable=False
                ),
                "member": pa.Column(
                    pa.String,
                    nullable=True
                )
            },
            strict="filter",
            coerce=True,
            add_missing_columns=True,
        )
    if type == "pyspark":
        return ps.DataFrameSchema(
            columns={
                "reference_time": ps.Column(
                    T.TimestampNTZType(),
                    nullable=True
                ),
                "value_time": ps.Column(
                    T.TimestampNTZType(),
                    nullable=False
                ),
                "value": ps.Column(
                    pyspark_value_type,
                    nullable=False,
                    coerce=True,
                ),
                "variable_name": ps.Column(
                    T.StringType(),
                    ps.Check.isin(variable_names),
                    nullable=False
                ),
                "configuration_name": ps.Column(
                    T.StringType(),
                    ps.Check.isin(configuration_names),
                    nullable=False
                ),
                "unit_name": ps.Column(
                    T.StringType(),
                    ps.Check.isin(unit_names),
                    nullable=False
                ),
                "location_id": ps.Column(
                    T.StringType(),
                    ps.Check.isin(location_ids),
                    nullable=False
                ),
                "member": ps.Column(
                    T.StringType(),
                    nullable=True,

                )
            },
            strict=True,
            coerce=True,
        )

def joined_timeseries_schema(
    type: str = "pyspark",
) -> ps.DataFrameSchema:
    if type == "pandas":
        return pa.DataFrameSchema(
            columns={
                "reference_time": pa.Column(
                    pa.DateTime,
                    parsers=pa.Parser(format_datetime64),
                    nullable=True
                ),
                "value_time": pa.Column(
                    pa.DateTime,
                    parsers=pa.Parser(format_datetime64),
                    nullable=False
                ),
                "primary_value": pa.Column(
                    pandas_value_type,
                    nullable=False,
                ),
                "secondary_value": pa.Column(
                    pandas_value_type,
                    nullable=False,
                ),
                "variable_name": pa.Column(
                    pa.String,
                    nullable=False
                ),
                "configuration_name": pa.Column(
                    pa.String,
                    nullable=False
                ),
                "unit_name": pa.Column(
                    pa.String,
                    nullable=False
                ),
                "primary_location_id": pa.Column(
                    pa.String,
                    nullable=False
                ),
                "secondary_location_id": pa.Column(
                    pa.String,
                    nullable=False
                ),
                "member": pa.Column(
                    pa.String,
                    nullable=True
                )
            }
        )
    if type == "pyspark":
        return ps.DataFrameSchema(
            columns={
                "reference_time": ps.Column(
                    T.TimestampNTZType(),
                    nullable=True
                ),
                "value_time": ps.Column(
                    T.TimestampNTZType(),
                    nullable=False
                ),
                "primary_value": ps.Column(
                    pyspark_value_type,
                    nullable=False,
                    coerce=True,
                ),
                "secondary_value": ps.Column(
                    pyspark_value_type,
                    nullable=False,
                    coerce=True,
                ),
                "variable_name": ps.Column(
                    T.StringType(),
                    nullable=False
                ),
                "configuration_name": ps.Column(
                    T.StringType(),
                    nullable=False
                ),
                "unit_name": ps.Column(
                    T.StringType(),
                    nullable=False
                ),
                "primary_location_id": ps.Column(
                    T.StringType(),
                    nullable=False
                ),
                "secondary_location_id": ps.Column(
                    T.StringType(),
                    nullable=False
                ),
                "member": ps.Column(
                    T.StringType(),
                    nullable=True
                )
            },
            coerce=True,
        )


"""Module to fetch test NWM streamflow data for the user guide."""

import requests


files = [
    {
        "name": "usgs_at_radford_location.parquet",
        "url": "https://github.com/RTIInternational/teehr/raw/refs/heads/main/tests/data/fetch_nwm_streamflow/usgs_at_radford_location.parquet"
    },
    {
        "name": "usgs_timeseries.parquet",
        "url": "https://github.com/RTIInternational/teehr/raw/refs/heads/main/tests/data/fetch_nwm_streamflow/usgs_timeseries.parquet"
    },
    {
        "name": "nwm_forecasts.parquet",
        "url": "https://github.com/RTIInternational/teehr/raw/refs/heads/main/tests/data/fetch_nwm_streamflow/nwm_forecasts.parquet"
    },
    {
        "name": "location_crosswalks.parquet",
        "url": "https://github.com/RTIInternational/teehr/raw/refs/heads/main/tests/data/fetch_nwm_streamflow/location_crosswalks.parquet"
    }
]


def fetch_and_save_file(url: str, destination: str) -> None:
    """Fetch a file from a URL and save it to a destination."""
    response = requests.get(url)
    response.raise_for_status()  # Ensure we notice bad responses

    with open(destination, 'wb') as f:
        f.write(response.content)


def list_files() -> list:
    """List the available files."""
    return [file["name"] for file in files]


def fetch_file(file_name: str, destination: str) -> None:
    """Fetch a file by name."""
    file = next(file for file in files if file["name"] == file_name)
    fetch_and_save_file(file["url"], destination)


from teehr.evaluation.tables.domain_table import DomainTable
from teehr.models.filters import AttributeFilter
from teehr.models.table_enums import AttributeFields
from teehr.models.pydantic_table_models import Attribute
import teehr.models.pandera_dataframe_schemas as schemas
from typing import List, Union
from teehr.utils.utils import to_path_or_s3path


class AttributeTable(DomainTable):
    """Access methods to attributes table."""

    def __init__(self, ev):
        """Initialize class."""
        super().__init__(ev)
        self.name = "attributes"
        self.dir = to_path_or_s3path(ev.dataset_dir, self.name)
        self.filter_model = AttributeFilter
        self.schema_func = schemas.attribute_schema
        self.unique_column_set = ["name"]

    def field_enum(self) -> AttributeFields:
        """Get the attribute fields enum."""
        fields = self._get_schema("pandas").columns.keys()
        return AttributeFields(
            "AttributeFields",
            {field: field for field in fields}
        )

    def add(
        self,
        attribute: Union[Attribute, List[Attribute]]
    ):
        """Add an attribute to the evaluation.

        Parameters
        ----------
        attribute : Union[Attribute, List[Attribute]]
            The attribute domain to add.

        Example
        -------
        >>> from teehr.models.domain_tables import Attribute
        >>> attribute = Attribute(
        >>>     name="drainage_area",
        >>>     type="continuous",
        >>>     description="Drainage area in square kilometers"
        >>> )
        >>> ev.load.add_attribute(attribute)
        """
        self._add(attribute)


"""Utility functions for the preprocessing."""
import geopandas as gpd
import pandas as pd
import xarray as xr
from pathlib import Path
from typing import Union, List
import logging
import shutil
from xml.dom import minidom
from pyspark.sql import DataFrame
import pyspark.sql.functions as F


logger = logging.getLogger(__name__)


def add_or_replace_sdf_column_prefix(
    sdf: DataFrame,
    column_name: str,
    prefix: str,
    delimiter: str = "-",
) -> DataFrame:
    """Add or replace a column in a DataFrame with a prefix.

    Parameters
    ----------
    sdf : DataFrame
        The DataFrame to add or replace the column in.
    column_name : str
        The name of the column to add or replace the prefix.
    prefix : str
        The prefix to add or replace.
    """
    if len(
        sdf.select(column_name).first().asDict()[column_name].split("-")
    ) > 1:
        logger.debug(
            "ID already has a prefix. This will be replaced."
        )
        sdf = sdf.withColumn(
            column_name,
            F.concat(
                F.lit(prefix), F.lit("-"), F.split(F.col(column_name), delimiter).getItem(1)
            )
        )
    else:
        logger.debug(
            "ID does not have a prefix. This will be added."
        )
        sdf = sdf.withColumn(
            column_name,
            F.concat(F.lit(prefix), F.lit(delimiter), F.col(column_name))
        )
    return sdf


def read_spatial_file(
        filepath: Union[str, Path], **kwargs: str
) -> gpd.GeoDataFrame:
    """Load any supported geospatial file type into a gdf using GeoPandas."""
    logger.info(f"Reading spatial file {filepath}")
    try:
        gdf = gpd.read_file(filepath, **kwargs)
        return gdf
    except Exception:
        pass
    try:
        gdf = gpd.read_parquet(filepath, **kwargs)
        return gdf
    except Exception:
        pass
    try:
        gdf = gpd.read_feather(filepath, **kwargs)
        return gdf
    except Exception:
        logger.error(f"Unsupported file type {filepath}")
        raise Exception("Unsupported file type")


def validate_dataset_structure(dataset_filepath: Union[str, Path]) -> bool:
    """Validate the database structure."""
    if not Path(dataset_filepath).exists():
        return False

    if not Path(dataset_filepath).is_dir():
        return False

    subdirectories = [
        "attributes",
        "configurations",
        "joined_timeseries",
        "location_crosswalks",
        "location_attributes",
        "locations",
        "primary_timeseries",
        "secondary_timeseries",
        "units",
        "variables",
    ]
    for subdirectory in subdirectories:
        if not Path(dataset_filepath, subdirectory).exists():
            logger.error(f"Subdirectory {subdirectory} not found.")
            return False

    return True


def validate_constant_values_dict(
    constant_values_dict: dict,
    fields: List[str],
):
    """Validate the values of the constants."""
    constant_keys = constant_values_dict.keys()
    for key in constant_keys:
        if key not in fields:
            logger.error(
                f"Invalid constant value key: {key}"
                f"Valid keys are: {fields}"
            )
            raise ValueError(f"Invalid constant value key: {key}")

    return True


def merge_field_mappings(
        default_mapping: dict,
        custom_mapping: dict
) -> dict:
    """Merge the default mapping with the custom mapping.

    Parameters
    ----------
    default_mapping : dict
        The default mapping.
        Format: {input_field: output_field}
    custom_mapping : dict
        The custom mapping.
        Format: {input_field: output_field}

    Returns
    -------
    dict
        The merged mapping.

    Merges based on output field names.
    """
    default_values = set(default_mapping.values())
    custom_values = set(custom_mapping.values())
    if not custom_values.issubset(default_values):
        logger.error(
            f"All custom values must be in default values: {default_values}"
        )
        raise ValueError("All custom values must be in default values")

    default_mapping = default_mapping.copy()
    custom_mapping = custom_mapping.copy()
    default_mapping = {v: k for k, v in default_mapping.items()}
    custom_mapping = {v: k for k, v in custom_mapping.items()}

    default_mapping.update(custom_mapping)
    mapping = {v: k for k, v in default_mapping.items()}
    return mapping


def copy_template_to(
        template_dir: Union[str, Path],
        destination_dir: Union[str, Path]
):
    """Copy the template directory to the destination directory."""
    template_dir = Path(template_dir)
    destination_dir = Path(destination_dir)

    for file in template_dir.glob('**/*'):
        if file.is_dir():
            destination_file = Path(
                destination_dir,
                file.relative_to(template_dir)
            )
            if not destination_file.parent.is_dir():
                destination_file.parent.mkdir(parents=True)
            logger.debug(f"Making directory {destination_file}")
            destination_file.mkdir()
        if file.is_file():
            destination_file = Path(
                destination_dir,
                file.relative_to(template_dir)
            )
            if not destination_file.parent.is_dir():
                destination_file.parent.mkdir(parents=True)
            logger.debug(f"Copying file {file} to {destination_file}")
            shutil.copy(file, destination_file)

    logger.debug(
        f"Renaming {destination_dir}/gitignore_template to .gitignore"
    )

    gitignore_text = Path(destination_dir, "gitignore_template")
    gitignore_text.rename(Path(destination_dir, ".gitignore"))


def read_and_convert_netcdf_to_df(
        in_filepath: Union[str, Path],
        field_mapping: dict,
        **kwargs
) -> pd.DataFrame:
    """Read a netcdf file and convert to pandas dataframe."""
    logger.debug(f"Reading and converting netcdf file {in_filepath}")
    with xr.open_dataset(in_filepath, **kwargs) as ds:
        # Get only the fields that are included in the field mapping.
        field_list = [field for field in field_mapping if field in ds]
        df = ds[field_list].to_dataframe()
    df.reset_index(inplace=True)
    return df


def _get_element_node_value(
    element: minidom.Element,
    tag_name: str
) -> str:
    """Get the node value of an element."""
    elements_list = element.getElementsByTagName(tag_name)
    if len(elements_list) == 0:
        logger.error(f"'{tag_name}' not found, unable to parse this file.")
        raise ValueError(
            f"'{tag_name}' not found, unable to parse this XML file."
        )
    elif len(elements_list) > 1:
        logger.error(
            f"More than one entry for'{tag_name}', unable to parse this file."
        )
        raise ValueError(
            f"More than one entry for'{tag_name}', unable to parse this file."
        )
    else:
        node_value = elements_list[0].firstChild.nodeValue
        return node_value


def _get_element_attribute_value(
    element: minidom.Element,
    tag_name: str,
    attribute_name: str
) -> str:
    """Get the attribute value of an element."""
    elements_list = element.getElementsByTagName(tag_name)
    if len(elements_list) == 0:
        logger.error(f"'{tag_name}' not found, unable to parse this file.")
        raise ValueError(
            f"'{tag_name}' not found, unable to parse this XML file."
        )
    elif len(elements_list) > 1:
        logger.error(
            f"More than one entry for'{tag_name}', unable to parse this file."
        )
        raise ValueError(
            f"More than one entry for'{tag_name}', unable to parse this file."
        )
    else:
        attr_value = elements_list[0].getAttribute(attribute_name)
        if len(attr_value) == 0:
            logger.error(
                f"Attribute '{attribute_name}' not found for '{tag_name}'."
            )
            raise ValueError(
                f"Attribute '{attribute_name}' not found for '{tag_name}'."
            )
        return attr_value


def read_and_convert_xml_to_df(
    in_filepath: Union[str, Path],
    field_mapping: dict,
) -> pd.DataFrame:
    """Read an xml file and convert to pandas dataframe."""
    logger.debug(f"Reading and converting xml file {in_filepath}")

    inv_field_mapping = {v: k for k, v in field_mapping.items()}

    location_id_kw = inv_field_mapping["location_id"]
    variable_name_kw = inv_field_mapping["variable_name"]
    reference_time_kw = inv_field_mapping["reference_time"]
    unit_name_kw = inv_field_mapping["unit_name"]
    member_kw = inv_field_mapping["member"]
    configuration_kw = inv_field_mapping["configuration_name"]

    fews_xml = minidom.parse(str(in_filepath))

    utc_offset = pd.Timedelta(
        float(
            fews_xml.getElementsByTagName("timeZone")[0].firstChild.nodeValue
        ),
        format="H"
    )
    series = fews_xml.getElementsByTagName("series")

    timeseries_list = []
    for member in series:
        location_id = _get_element_node_value(member, location_id_kw)
        variable_name = _get_element_node_value(member, variable_name_kw)
        configuration = _get_element_node_value(member, configuration_kw)
        unit_name = _get_element_node_value(member, unit_name_kw)
        ensemble_member = _get_element_node_value(member, member_kw)

        forecastDate = _get_element_attribute_value(member, reference_time_kw, "date")
        forecastTime = _get_element_attribute_value(member, reference_time_kw, "time")

        reference_time = pd.to_datetime(
            forecastDate + " " + forecastTime
        ) + utc_offset

        events = member.getElementsByTagName("event")
        timeseries_data = []
        for event in events:
            event_date = event.getAttribute("date")
            event_time = event.getAttribute("time")
            event_value = event.getAttribute("value")
            value_time = pd.to_datetime(
                event_date + " " + event_time
            ) + utc_offset
            timeseries_data.append({
                "value": event_value,
                "value_time": value_time
            })

        timeseries_df = pd.DataFrame(timeseries_data)
        timeseries_df["reference_time"] = reference_time
        timeseries_df["unit_name"] = unit_name
        timeseries_df["variable_name"] = variable_name
        timeseries_df["location_id"] = location_id
        timeseries_df["member"] = ensemble_member
        timeseries_df["configuration_name"] = configuration

        timeseries_list.append(timeseries_df)

    return pd.concat(timeseries_list)


def validate_input_is_parquet(
        in_path: Union[str, Path]
) -> None:
    """
    Validate that the input is parquet format.

    Check that either the file is a parquet file or the directory
    contains parquet files.
    """
    in_path = Path(in_path)
    if in_path.is_dir():
        if len(list(in_path.glob("**/*.parquet"))) == 0:
            logger.error("No parquet files found in the directory.")
            raise ValueError("No parquet files found in the directory.")
    else:
        if not in_path.suffix.endswith(".parquet"):
            logger.error("Input file must be a parquet file.")
            raise ValueError("Input file must be a parquet file.")


def validate_input_is_csv(
        in_path: Union[str, Path]
) -> None:
    """
    Validate that the input is csv format.

    Check that either the file is a csv file or the directory
    contains csv files.
    """
    in_path = Path(in_path)
    if in_path.is_dir():
        if len(list(in_path.glob("**/*.csv"))) == 0:
            logger.error("No csv files found in the directory.")
            raise ValueError("No csv files found in the directory.")
    else:
        if not in_path.suffix.endswith(".csv"):
            logger.error("Input file must be a csv file.")
            raise ValueError("Input file must be a csv file.")


def validate_input_is_netcdf(
        in_path: Union[str, Path]
) -> None:
    """
    Validate that the input is netcdf format.

    Check that either the file is a netcdf file or the directory
    contains netcdf files.
    """
    in_path = Path(in_path)
    if in_path.is_dir():
        if len(list(in_path.glob("**/*.nc"))) == 0:
            logger.error("No netcdf files found in the directory.")
            raise ValueError("No netcdf files found in the directory.")
    else:
        if not in_path.suffix.endswith(".nc"):
            logger.error("Input file must be a netcdf file.")
            raise ValueError("Input file must be a netcdf file.")


def validate_input_is_xml(
        in_path: Union[str, Path]
) -> None:
    """
    Validate that the input is xml format.

    Check that either the file is a xml file or the directory
    contains xml files.
    """
    in_path = Path(in_path)
    if in_path.is_dir():
        if len(list(in_path.glob("**/*.xml"))) == 0:
            logger.error("No xml files found in the directory.")
            raise ValueError("No xml files found in the directory.")
    else:
        if not in_path.suffix.endswith(".xml"):
            logger.error("Input file must be an xml file.")
            raise ValueError("Input file must be an xml file.")



"""Functions for adding user defined fields to joined timeseries data.

This file gets dynamically imported to the TEEHR
create_joined_timeseries_dataset() function and is used to add user
defined fields when a user runs ev.create_joined_timeseries().

Users can modify this script to add or remove user defined fields to
the joined timeseries data.

WARNING: Do not change the name of this file or the functions it contains.
"""

from pyspark.sql import DataFrame
import logging
from teehr import RowLevelCalculatedFields as rcf
from teehr import TimeseriesAwareCalculatedFields as tcf

logger = logging.getLogger(__name__)


def add_user_defined_fields(
    joined_df: DataFrame
):
    """Add user defined fields to the joined timeseries data.

    Do not change the name of this function.

    Parameters
    ----------
    joined_timeseries : JoinedTimeseriesTable

    Returns
    -------
    DataFrame
        The joined timeseries data with user defined fields.
    """
    logger.info("Adding user defined fields")

    # Add a month field to the joined timeseries data
    logger.info("Adding month from date")

    month = rcf.Month()
    year = rcf.Year()
    water_year = rcf.WaterYear()
    # normalized_flow = rcf.NormalizedFlow()
    seasons = rcf.Seasons()

    cfs = [
        month,
        year,
        water_year,
        # normalized_flow,
        seasons
    ]

    for cf in cfs:
        joined_df = cf.apply_to(joined_df)

    # Return the joined timeseries data with user defined fields
    return joined_df


"""Module for generating metrics."""
from typing import Union, List

import pandas as pd
import geopandas as gpd
import pyspark.sql as ps
from teehr.models.filters import (
    JoinedTimeseriesFilter
)
from teehr.models.metrics.basemodels import MetricsBasemodel
from teehr.models.calculated_fields.base import CalculatedFieldBaseModel
from teehr.models.table_enums import (
    JoinedTimeseriesFields
)
from teehr.querying.filter_format import validate_and_apply_filters
from teehr.querying.metric_format import apply_aggregation_metrics
from teehr.querying.utils import (
    order_df,
    group_df,
    join_geometry
)

import logging

logger = logging.getLogger(__name__)


class Metrics:
    """Component class for calculating metrics."""

    def __init__(self, ev) -> None:
        """Initialize the Metrics class."""
        self.spark = ev.spark
        self.dataset_dir = ev.dataset_dir
        self.locations = ev.locations
        self.joined_timeseries = ev.joined_timeseries
        self.df = self.joined_timeseries.to_sdf()

    def query(
        self,
        filters: Union[
            str, dict, JoinedTimeseriesFilter,
            List[Union[str, dict, JoinedTimeseriesFilter]]
        ] = None,
        order_by: Union[
            str, JoinedTimeseriesFields,
            List[Union[str, JoinedTimeseriesFields]]
        ] = None,
        group_by: Union[
            str, JoinedTimeseriesFields,
            List[Union[str, JoinedTimeseriesFields]]
        ] = None,
        include_metrics: Union[
            List[MetricsBasemodel],
            str
        ] = None
    ):
        """Perform a query on the dataset joined timeseries table.

        Parameters
        ----------
        filters : Union[str, dict, JoinedTimeseriesFilter, List[Union[str, dict, JoinedTimeseriesFilter]]], optional
            The filters to apply to the query, by default None
        order_by : Union[str, JoinedTimeseriesFields, List[Union[str, JoinedTimeseriesFields]]], optional
            The fields to order the query by, by default None
        group_by : Union[str, JoinedTimeseriesFields, List[Union[str, JoinedTimeseriesFields]]], optional
            The fields to group the query by, by default None
        include_metrics : Union[List[MetricsBasemodel], str], optional
            The metrics to include in the query, by default None

        Examples
        --------

        >>> import teehr
        >>> ev = teehr.Evaluation()

        Define some metrics, optionally including an available bootstrapping
        method. (:func:`Metric Models <teehr.models.metrics.metric_models>`).

        >>> import teehr.Metrics as m
        >>> import teehr.Bootstrappers as b

        Define a Circular Block bootstrapper.
        (:func:`Bootstrap Models <teehr.models.metrics.bootstrap_models>`).

        >>> boot = b.CircularBlock(
        >>>     seed=40,
        >>>     block_size=100,
        >>>     quantiles=None,
        >>>     reps=500
        >>> )

        Include the bootstrap model in the metric definition(s), along with other
        optional arguments.

        >>> kge = m.KlingGuptaEfficiency(bootstrap=boot)
        >>> primary_avg = m.Average(
        >>>     transform="log",
        >>>     output_field_name="primary_avg",
        >>>     input_field_names=["primary_value"]
        >>> )
        >>> mvtd = m.MaxValueTimeDelta(input_field_names=["secondary_value"])
        >>> pmvt = m.MaxValueTime(input_field_names=["secondary_value"])

        >>> include_metrics = [pmvt, mvtd, primary_avg, kge]

        Get the currently available fields to use in the query.

        >>> flds = eval.joined_timeseries.field_enum()

        Define some filters.

        >>> filters = [
        >>>     JoinedTimeseriesFilter(
        >>>         column=flds.primary_location_id,
        >>>         operator=ops.eq,
        >>>         value="gage-A"
        >>>     )
        >>> ]

        Perform the query, returning the results as a GeoPandas DataFrame.

        >>> metrics_df = eval.metrics.query(
        >>>     include_metrics=include_metrics,
        >>>     group_by=[flds.primary_location_id],
        >>>     order_by=[flds.primary_location_id],
        >>>     filters=filters,
        >>> ).to_geopandas()
        """ # noqa
        logger.info("Calculating performance metrics.")
        if filters is not None:
            logger.debug("Applying filters to the metrics query.")
            self.df = validate_and_apply_filters(
                sdf=self.df,
                filter_model=self.joined_timeseries.filter_model,
                filters=filters,
                fields_enum=self.joined_timeseries.field_enum(),
                validate=False
            )

        logger.debug(f"Grouping the metrics query {group_by}.")
        self.df = group_df(self.df, group_by)

        self.df = apply_aggregation_metrics(
            self.df,
            include_metrics
        )

        if order_by is not None:
            logger.debug(f"Ordering the metrics by: {order_by}.")
            self.df = order_df(self.df, order_by)

        return self

    def to_pandas(self) -> pd.DataFrame:
        """Convert the DataFrame to a Pandas DataFrame."""
        df = self.df.toPandas()
        df.attrs['table_type'] = 'metrics'
        return df

    def to_geopandas(self) -> gpd.GeoDataFrame:
        """Convert the DataFrame to a GeoPandas DataFrame."""
        if "primary_location_id" not in self.df.columns:
            err_msg = "The primary_location_id field must be included in " \
                      "the group_by to include geometry."
            logger.error(err_msg)
            raise ValueError(err_msg)
        return join_geometry(
            self.df,
            self.locations.to_sdf(),
            "primary_location_id"
        )

    def to_sdf(self) -> ps.DataFrame:
        """Return the Spark DataFrame."""
        return self.df

    def add_calculated_fields(self, cfs: Union[CalculatedFieldBaseModel, List[CalculatedFieldBaseModel]]):
            """Add calculated fields to the joined timeseries DataFrame before running metrics.

            Parameters
            ----------
            cfs : Union[CalculatedFieldBaseModel, List[CalculatedFieldBaseModel]]
                The CFs to apply to the DataFrame.

            Returns
            -------
            self
                The Metrics object with the CFs applied to the DataFrame.

            Examples
            --------
            >>> import teehr
            >>> from teehr import RowLevelCalculatedFields as rcf
            >>> ev.join_timeseries.add_calculated_fields([
            >>>     rcf.Month()
            >>> ]).write()
            """
            # self._check_load_table()

            if not isinstance(cfs, List):
                cfs = [cfs]

            for cf in cfs:
                self.df = cf.apply_to(self.df)

            return self